# Stubs for networkx.algorithms.centrality.dispersion (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def dispersion(
    G,
    u: Optional[Any] = ...,
    v: Optional[Any] = ...,
    normalized: bool = ...,
    alpha: float = ...,
    b: float = ...,
    c: float = ...,
): ...



# Stubs for networkx.algorithms.approximation.vertex_cover (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def min_weighted_vertex_cover(G, weight: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.connectivity (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from .connectivity import *
from .cuts import *
from .disjoint_paths import *
from .edge_augmentation import *
from .edge_kcomponents import *
from .kcomponents import *
from .kcutsets import *
from .stoerwagner import *
from .utils import *

# Names in __all__ with no definition:
#   EdgeComponentAuxGraph
#   all_node_cuts
#   all_pairs_node_connectivity
#   average_node_connectivity
#   bridge_components
#   build_auxiliary_edge_connectivity
#   build_auxiliary_node_connectivity
#   edge_connectivity
#   edge_disjoint_paths
#   is_k_edge_connected
#   is_locally_k_edge_connected
#   k_components
#   k_edge_augmentation
#   k_edge_components
#   k_edge_subgraphs
#   local_edge_connectivity
#   local_node_connectivity
#   minimum_edge_cut
#   minimum_node_cut
#   minimum_st_edge_cut
#   minimum_st_node_cut
#   node_connectivity
#   node_disjoint_paths
#   stoer_wagner



from _typeshed import Incomplete

logger: Incomplete
URI: Incomplete

class CacheController:
    cache: Incomplete
    cache_etags: Incomplete
    serializer: Incomplete
    cacheable_status_codes: Incomplete
    def __init__(
        self,
        cache: Incomplete | None = ...,
        cache_etags: bool = ...,
        serializer: Incomplete | None = ...,
        status_codes: Incomplete | None = ...,
    ) -> None: ...



# Stubs for networkx.algorithms.shortest_paths (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.algorithms.shortest_paths.astar import *
from networkx.algorithms.shortest_paths.dense import *
from networkx.algorithms.shortest_paths.generic import *
from networkx.algorithms.shortest_paths.unweighted import *
from networkx.algorithms.shortest_paths.weighted import *



# Stubs for networkx.generators.lattice (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def grid_2d_graph(m, n, periodic: bool = ..., create_using: Optional[Any] = ...): ...
def grid_graph(dim, periodic: bool = ...): ...
def hypercube_graph(n): ...
def triangular_lattice_graph(
    m,
    n,
    periodic: bool = ...,
    with_positions: bool = ...,
    create_using: Optional[Any] = ...,
): ...
def hexagonal_lattice_graph(
    m,
    n,
    periodic: bool = ...,
    with_positions: bool = ...,
    create_using: Optional[Any] = ...,
): ...



# Stubs for networkx.algorithms.operators.all (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

def union_all(graphs, rename: Any = ...): ...
def disjoint_union_all(graphs): ...
def compose_all(graphs): ...
def intersection_all(graphs): ...



"""Tests for --print-inputs-deps."""

import json
import os
from io import StringIO
from pathlib import Path

from cwltool.main import main
from cwltool.process import CWL_IANA

from .util import get_data


def test_input_deps() -> None:
    """Basic test of --print-input-deps with a job input file."""
    stream = StringIO()

    main(
        [
            "--print-input-deps",
            get_data("tests/wf/count-lines1-wf.cwl"),
            get_data("tests/wf/wc-job.json"),
        ],
        stdout=stream,
    )

    expected = {
        "class": "File",
        "location": "wc-job.json",
        "format": CWL_IANA,
        "secondaryFiles": [
            {
                "class": "File",
                "location": "whale.txt",
                "basename": "whale.txt",
                "nameroot": "whale",
                "nameext": ".txt",
            }
        ],
    }
    assert json.loads(stream.getvalue()) == expected


def test_input_deps_cmdline_opts() -> None:
    """Test of --print-input-deps with command line provided inputs."""
    stream = StringIO()

    main(
        [
            "--print-input-deps",
            get_data("tests/wf/count-lines1-wf.cwl"),
            "--file1",
            get_data("tests/wf/whale.txt"),
        ],
        stdout=stream,
    )
    expected = {
        "class": "File",
        "location": "",
        "format": CWL_IANA,
        "secondaryFiles": [
            {
                "class": "File",
                "location": "whale.txt",
                "basename": "whale.txt",
                "nameroot": "whale",
                "nameext": ".txt",
            }
        ],
    }
    assert json.loads(stream.getvalue()) == expected


def test_input_deps_cmdline_opts_relative_deps_cwd() -> None:
    """Test of --print-input-deps with command line provided inputs and relative-deps."""
    stream = StringIO()

    data_path = get_data("tests/wf/whale.txt")
    main(
        [
            "--print-input-deps",
            "--relative-deps",
            "cwd",
            get_data("tests/wf/count-lines1-wf.cwl"),
            "--file1",
            data_path,
        ],
        stdout=stream,
    )

    goal = {
        "class": "File",
        "location": "",
        "format": CWL_IANA,
        "secondaryFiles": [
            {
                "class": "File",
                "location": str(Path(os.path.relpath(data_path, os.path.curdir))),
                "basename": "whale.txt",
                "nameroot": "whale",
                "nameext": ".txt",
            }
        ],
    }
    assert json.loads(stream.getvalue()) == goal


def test_input_deps_secondary_files() -> None:
    """Affirm that secondaryFiles are also represented."""
    stream = StringIO()

    main(
        [
            "--print-input-deps",
            get_data("tests/input_deps/docker-array-secondaryfiles.cwl"),
            get_data("tests/input_deps/docker-array-secondaryfiles-job.json"),
        ],
        stdout=stream,
    )

    goal = {
        "class": "File",
        "location": "docker-array-secondaryfiles-job.json",
        "format": CWL_IANA,
        "secondaryFiles": [
            {
                "class": "File",
                "location": "ref.fasta",
                "basename": "ref.fasta",
                "secondaryFiles": [
                    {
                        "class": "File",
                        "location": "ref.fasta.fai",
                        "basename": "ref.fasta.fai",
                        "nameroot": "ref.fasta",
                        "nameext": ".fai",
                    }
                ],
                "nameroot": "ref",
                "nameext": ".fasta",
            },
            {
                "class": "File",
                "location": "ref2.fasta",
                "basename": "ref2.fasta",
                "secondaryFiles": [
                    {
                        "class": "File",
                        "location": "ref2.fasta.fai",
                        "basename": "ref2.fasta.fai",
                        "nameroot": "ref2.fasta",
                        "nameext": ".fai",
                    }
                ],
                "nameroot": "ref2",
                "nameext": ".fasta",
            },
        ],
    }

    assert json.loads(stream.getvalue()) == goal



# Stubs for networkx.readwrite.gexf (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def write_gexf(G, path, encoding: str = ..., prettyprint: bool = ..., version: str = ...): ...
def generate_gexf(G, encoding: str = ..., prettyprint: bool = ..., version: str = ...): ...
def read_gexf(path, node_type: Optional[Any] = ..., relabel: bool = ..., version: str = ...): ...

class GEXF:
    versions: Any = ...
    d: Any = ...
    types: Any = ...
    blurb: Any = ...
    xml_type: Any = ...
    python_type: Any = ...
    convert_bool: Any = ...
    NS_GEXF: Any = ...
    NS_VIZ: Any = ...
    NS_XSI: Any = ...
    SCHEMALOCATION: Any = ...
    VERSION: Any = ...
    version: Any = ...
    def set_version(self, version): ...

class GEXFWriter(GEXF):
    prettyprint: Any = ...
    encoding: Any = ...
    xml: Any = ...
    edge_id: Any = ...
    attr_id: Any = ...
    attr: Any = ...
    def __init__(
        self,
        graph: Optional[Any] = ...,
        encoding: str = ...,
        prettyprint: bool = ...,
        version: str = ...,
    ) -> None: ...
    graph_element: Any = ...
    def add_graph(self, G): ...
    def add_meta(self, G, graph_element): ...
    def add_nodes(self, G, graph_element): ...
    def add_edges(self, G, graph_element): ...
    def add_attributes(self, node_or_edge, xml_obj, data, default): ...
    def get_attr_id(self, title, attr_type, edge_or_node, default, mode): ...
    def add_viz(self, element, node_data): ...
    def add_parents(self, node_element, node_data): ...
    def add_slices(self, node_or_edge_element, node_or_edge_data): ...
    def add_spells(self, node_or_edge_element, node_or_edge_data): ...
    def alter_graph_mode_timeformat(self, start_or_end): ...
    def write(self, fh): ...
    def indent(self, elem, level: int = ...): ...

class GEXFReader(GEXF):
    node_type: Any = ...
    simple_graph: bool = ...
    def __init__(self, node_type: Optional[Any] = ..., version: str = ...) -> None: ...
    xml: Any = ...
    def __call__(self, stream): ...
    timeformat: Any = ...
    def make_graph(self, graph_xml): ...
    def add_node(self, G, node_xml, node_attr, node_pid: Optional[Any] = ...): ...
    def add_start_end(self, data, xml): ...
    def add_viz(self, data, node_xml): ...
    def add_parents(self, data, node_xml): ...
    def add_slices(self, data, node_or_edge_xml): ...
    def add_spells(self, data, node_or_edge_xml): ...
    def add_edge(self, G, edge_element, edge_attr): ...
    def decode_attr_elements(self, gexf_keys, obj_xml): ...
    def find_gexf_attributes(self, attributes_element): ...

def relabel_gexf_graph(G): ...



# Stubs for networkx.readwrite.json_graph.jit (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def jit_graph(data, create_using: Optional[Any] = ...): ...
def jit_data(G, indent: Optional[Any] = ...): ...



"""This module handles resolution of SoftwareRequirement hints.

This is accomplished mainly by adapting cwltool internals to galaxy-tool-util's
concept of "dependencies". Despite the name, galaxy-tool-util is a light weight
library that can be used to map SoftwareRequirements in all sorts of ways -
Homebrew, Conda, custom scripts, environment modules. We'd be happy to find
ways to adapt new packages managers and such as well.
"""

import argparse
import os
import string
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    MutableMapping,
    MutableSequence,
    Optional,
    Union,
    cast,
)

from .utils import HasReqsHints

if TYPE_CHECKING:
    from .builder import Builder

try:
    from galaxy.tool_util import deps
    from galaxy.tool_util.deps.requirements import ToolRequirement, ToolRequirements
except ImportError:
    ToolRequirement = None  # type: ignore
    ToolRequirements = None  # type: ignore
    deps = None  # type: ignore


SOFTWARE_REQUIREMENTS_ENABLED = deps is not None

COMMAND_WITH_DEPENDENCIES_TEMPLATE = string.Template(
    """#!/bin/bash
cat > modify_environment.bash <<'EOF'
$handle_dependencies
# First try env -0
if ! env -0 > "output_environment.dat" 2> /dev/null; then
    # If that fails, use the python script.
    # In some circumstances (see PEP 538) this will the add LC_CTYPE env var.
    python3 "env_to_stdout.py" > "output_environment.dat"
fi
EOF
python3 "run_job.py" "job.json" "modify_environment.bash"
"""
)


class DependenciesConfiguration:
    """Dependency configuration class, for RuntimeContext.job_script_provider."""

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize."""
        conf_file = getattr(args, "beta_dependency_resolvers_configuration", None)
        tool_dependency_dir = getattr(args, "beta_dependencies_directory", None)
        conda_dependencies = getattr(args, "beta_conda_dependencies", None)
        if conf_file is not None and os.path.exists(conf_file):
            self.use_tool_dependencies = True
            if tool_dependency_dir is None:
                tool_dependency_dir = os.path.abspath(os.path.dirname(conf_file))
            self.tool_dependency_dir = tool_dependency_dir
            self.dependency_resolvers_config_file = os.path.abspath(conf_file)
        elif conda_dependencies is not None:
            if tool_dependency_dir is None:
                tool_dependency_dir = os.path.abspath("./cwltool_deps")
            self.tool_dependency_dir = tool_dependency_dir
            self.use_tool_dependencies = True
            self.dependency_resolvers_config_file = None
        else:
            self.use_tool_dependencies = False

    def build_job_script(self, builder: "Builder", command: List[str]) -> str:
        ensure_galaxy_lib_available()
        resolution_config_dict = {
            "use": self.use_tool_dependencies,
            "default_base_path": self.tool_dependency_dir,
        }
        app_config = {
            "conda_auto_install": True,
            "conda_auto_init": True,
        }
        tool_dependency_manager: "deps.DependencyManager" = deps.build_dependency_manager(
            app_config_dict=app_config,
            resolution_config_dict=resolution_config_dict,
            conf_file=self.dependency_resolvers_config_file,
        )
        dependencies = get_dependencies(builder)
        handle_dependencies = ""  # str
        if dependencies:
            handle_dependencies = "\n".join(
                tool_dependency_manager.dependency_shell_commands(
                    dependencies, job_directory=builder.tmpdir
                )
            )

        template_kwds: Dict[str, str] = dict(handle_dependencies=handle_dependencies)
        job_script = COMMAND_WITH_DEPENDENCIES_TEMPLATE.substitute(template_kwds)
        return job_script


def get_dependencies(builder: HasReqsHints) -> ToolRequirements:
    (software_requirement, _) = builder.get_requirement("SoftwareRequirement")
    dependencies: List["ToolRequirement"] = []
    if software_requirement and software_requirement.get("packages"):
        packages = cast(
            MutableSequence[MutableMapping[str, Union[str, MutableSequence[str]]]],
            software_requirement.get("packages"),
        )
        for package in packages:
            version = package.get("version", None)
            if isinstance(version, MutableSequence):
                if version:
                    version = version[0]
                else:
                    version = None
            specs = [{"uri": s} for s in package.get("specs", [])]
            dependencies.append(
                ToolRequirement.from_dict(
                    dict(
                        name=cast(str, package["package"]).split("#")[-1],
                        version=version,
                        type="package",
                        specs=specs,
                    )
                )
            )

    return ToolRequirements.from_list(dependencies)


def get_container_from_software_requirements(
    use_biocontainers: bool, builder: HasReqsHints
) -> Optional[str]:
    if use_biocontainers:
        ensure_galaxy_lib_available()
        from galaxy.tool_util.deps.container_classes import DOCKER_CONTAINER_TYPE
        from galaxy.tool_util.deps.containers import ContainerRegistry
        from galaxy.tool_util.deps.dependencies import AppInfo, ToolInfo

        app_info: AppInfo = AppInfo(
            involucro_auto_init=True,
            enable_mulled_containers=True,
            container_image_cache_path=".",
        )
        container_registry: ContainerRegistry = ContainerRegistry(app_info)
        requirements = get_dependencies(builder)
        tool_info: ToolInfo = ToolInfo(requirements=requirements)
        container_description = container_registry.find_best_container_description(
            [DOCKER_CONTAINER_TYPE], tool_info
        )
        if container_description:
            return cast(Optional[str], container_description.identifier)

    return None


def ensure_galaxy_lib_available() -> None:
    if not SOFTWARE_REQUIREMENTS_ENABLED:
        raise Exception(
            "Optional Python library galaxy-tool-util not available, it is required for this configuration."
        )



# Stubs for networkx.algorithms.distance_regular (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def is_distance_regular(G): ...
def global_parameters(b, c): ...
def intersection_array(G): ...
def is_strongly_regular(G): ...



# Stubs for networkx.utils.union_find (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

class UnionFind:
    parents: Any = ...
    weights: Any = ...
    def __init__(self, elements: Optional[Any] = ...) -> None: ...
    def __getitem__(self, object): ...
    def __iter__(self): ...
    def to_sets(self): ...
    def union(self, *objects): ...



# Stubs for networkx.generators.stochastic (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def stochastic_graph(G, copy: bool = ..., weight: str = ...): ...



# Stubs for networkx.generators.duplication (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def partial_duplication_graph(N, n, p, q, seed: Optional[Any] = ...): ...
def duplication_divergence_graph(n, p, seed: Optional[Any] = ...): ...



"""Test user experience running on MS Windows."""

import os

import pytest

from cwltool import main

# Can't be just "import cwltool ; … cwltool.main.windows_check()"
# needs a direct import to avoid path traversal after os.name is set to "nt"


def test_windows_warning(monkeypatch: pytest.MonkeyPatch) -> None:
    """Confirm that the windows warning is given."""
    with pytest.warns(UserWarning, match=r"Windows Subsystem for Linux 2"):
        # would normally just use the MonkeyPatch object directly
        # but if we don't use a context then os.name being "nt" causes problems
        # for pytest on non-Windows systems. So the context unravels the change
        # to os.name quickly, and then pytest will check for the desired warning
        with monkeypatch.context() as m:
            m.setattr(os, "name", "nt")
            main.windows_check()



# Stubs for networkx.algorithms.bipartite.generators (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def complete_bipartite_graph(n1, n2, create_using: Optional[Any] = ...): ...
def configuration_model(
    aseq, bseq, create_using: Optional[Any] = ..., seed: Optional[Any] = ...
): ...
def havel_hakimi_graph(aseq, bseq, create_using: Optional[Any] = ...): ...
def reverse_havel_hakimi_graph(aseq, bseq, create_using: Optional[Any] = ...): ...
def alternating_havel_hakimi_graph(aseq, bseq, create_using: Optional[Any] = ...): ...
def preferential_attachment_graph(
    aseq, p, create_using: Optional[Any] = ..., seed: Optional[Any] = ...
): ...
def random_graph(n, m, p, seed: Optional[Any] = ..., directed: bool = ...): ...
def gnmk_random_graph(n, m, k, seed: Optional[Any] = ..., directed: bool = ...): ...



# Stubs for networkx.algorithms.approximation.dominating_set (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def min_weighted_dominating_set(G, weight: Optional[Any] = ...): ...
def min_edge_dominating_set(G): ...



# Stubs for networkx.relabel (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def relabel_nodes(G, mapping, copy: bool = ...): ...
def convert_node_labels_to_integers(
    G, first_label: int = ..., ordering: str = ..., label_attribute: Optional[Any] = ...
): ...



# Stubs for networkx.classes.filters (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

def no_filter(*items): ...
def hide_nodes(nodes): ...
def hide_diedges(edges): ...
def hide_edges(edges): ...
def hide_multidiedges(edges): ...
def hide_multiedges(edges): ...

class show_nodes:
    nodes: Any = ...
    def __init__(self, nodes) -> None: ...
    def __call__(self, node): ...

def show_diedges(edges): ...
def show_edges(edges): ...
def show_multidiedges(edges): ...
def show_multiedges(edges): ...



# Stubs for networkx.algorithms.covering (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def min_edge_cover(G, matching_algorithm: Optional[Any] = ...): ...
def is_edge_cover(G, cover): ...



# Stubs for networkx.classes.graph (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

class Graph:
    node_dict_factory: Any = ...
    adjlist_outer_dict_factory: Any = ...
    adjlist_inner_dict_factory: Any = ...
    edge_attr_dict_factory: Any = ...
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, incoming_graph_data: Optional[Any] = ..., **attr) -> None: ...
    @property
    def adj(self): ...
    @property
    def name(self): ...
    @name.setter
    def name(self, s): ...
    def __iter__(self): ...
    def __contains__(self, n): ...
    def __len__(self): ...
    def __getitem__(self, n): ...
    def add_node(self, node_for_adding, **attr): ...
    def add_nodes_from(self, nodes_for_adding, **attr): ...
    def remove_node(self, n): ...
    def remove_nodes_from(self, nodes): ...
    @property
    def nodes(self): ...
    node: Any = ...
    def add_path(self, nodes, **attr): ...
    def add_cycle(self, nodes, **attr): ...
    def add_star(self, nodes, **attr): ...
    def nodes_with_selfloops(self): ...
    def number_of_selfloops(self): ...
    def selfloop_edges(self, data: bool = ..., keys: bool = ..., default: Optional[Any] = ...): ...
    def number_of_nodes(self): ...
    def order(self): ...
    def has_node(self, n): ...
    def add_edge(self, u_of_edge, v_of_edge, **attr): ...
    def add_edges_from(self, ebunch_to_add, **attr): ...
    def add_weighted_edges_from(self, ebunch_to_add, weight: str = ..., **attr): ...
    def remove_edge(self, u, v): ...
    def remove_edges_from(self, ebunch): ...
    def has_edge(self, u, v): ...
    def neighbors(self, n): ...
    @property
    def edges(self): ...
    def get_edge_data(self, u, v, default: Optional[Any] = ...): ...
    def adjacency(self): ...
    @property
    def degree(self): ...
    def clear(self): ...
    def is_multigraph(self): ...
    def is_directed(self): ...
    def fresh_copy(self): ...
    def copy(self, as_view: bool = ...): ...
    def to_directed(self, as_view: bool = ...): ...
    def to_undirected(self, as_view: bool = ...): ...
    def subgraph(self, nodes): ...
    def edge_subgraph(self, edges): ...
    def size(self, weight: Optional[Any] = ...): ...
    def number_of_edges(self, u: Optional[Any] = ..., v: Optional[Any] = ...): ...
    def nbunch_iter(self, nbunch: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.cuts (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def cut_size(G, S, T: Optional[Any] = ..., weight: Optional[Any] = ...): ...
def volume(G, S, weight: Optional[Any] = ...): ...
def normalized_cut_size(G, S, T: Optional[Any] = ..., weight: Optional[Any] = ...): ...
def conductance(G, S, T: Optional[Any] = ..., weight: Optional[Any] = ...): ...
def edge_expansion(G, S, T: Optional[Any] = ..., weight: Optional[Any] = ...): ...
def mixing_expansion(G, S, T: Optional[Any] = ..., weight: Optional[Any] = ...): ...
def node_expansion(G, S): ...
def boundary_expansion(G, S): ...



# Stubs for networkx.classes.multigraph (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from networkx.classes.graph import Graph

class MultiGraph(Graph):
    edge_key_dict_factory: Any = ...
    def __init__(self, incoming_graph_data: Optional[Any] = ..., **attr) -> None: ...
    @property
    def adj(self): ...
    def new_edge_key(self, u, v): ...
    def add_edge(self, u_for_edge, v_for_edge, key: Optional[Any] = ..., **attr): ...
    def add_edges_from(self, ebunch_to_add, **attr): ...
    def remove_edge(self, u, v, key: Optional[Any] = ...): ...
    def remove_edges_from(self, ebunch): ...
    def has_edge(self, u, v, key: Optional[Any] = ...): ...
    @property
    def edges(self): ...
    def get_edge_data(self, u, v, key: Optional[Any] = ..., default: Optional[Any] = ...): ...
    @property
    def degree(self): ...
    def is_multigraph(self): ...
    def is_directed(self): ...
    def fresh_copy(self): ...
    def copy(self, as_view: bool = ...): ...
    def to_directed(self, as_view: bool = ...): ...
    def to_undirected(self, as_view: bool = ...): ...
    def subgraph(self, nodes): ...
    def number_of_edges(self, u: Optional[Any] = ..., v: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.centrality.betweenness (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def betweenness_centrality(
    G,
    k: Optional[Any] = ...,
    normalized: bool = ...,
    weight: Optional[Any] = ...,
    endpoints: bool = ...,
    seed: Optional[Any] = ...,
): ...
def edge_betweenness_centrality(
    G,
    k: Optional[Any] = ...,
    normalized: bool = ...,
    weight: Optional[Any] = ...,
    seed: Optional[Any] = ...,
): ...
def edge_betweenness(
    G,
    k: Optional[Any] = ...,
    normalized: bool = ...,
    weight: Optional[Any] = ...,
    seed: Optional[Any] = ...,
): ...



# Stubs for networkx.algorithms.coloring.greedy_coloring_with_interchange (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

class Node:
    node_id: Any = ...
    color: int = ...
    adj_list: Any = ...
    adj_color: Any = ...
    def __init__(self, node_id, n) -> None: ...
    def assign_color(self, adj_entry, color): ...
    def clear_color(self, adj_entry, color): ...
    def iter_neighbors(self): ...
    def iter_neighbors_color(self, color): ...

class AdjEntry:
    node_id: Any = ...
    next: Any = ...
    mate: Any = ...
    col_next: Any = ...
    col_prev: Any = ...
    def __init__(self, node_id) -> None: ...

def greedy_coloring_with_interchange(original_graph, nodes): ...



"""Loads a CWL document."""

import copy
import hashlib
import logging
import os
import re
import urllib
import uuid
from functools import partial
from typing import (
    Any,
    Dict,
    List,
    MutableMapping,
    MutableSequence,
    Optional,
    Tuple,
    Union,
    cast,
)

from cwl_utils.parser import cwl_v1_2, cwl_v1_2_utils
from schema_salad.exceptions import ValidationException
from schema_salad.fetcher import Fetcher
from schema_salad.ref_resolver import Loader, file_uri
from schema_salad.schema import validate_doc
from schema_salad.sourceline import SourceLine, cmap
from schema_salad.utils import (
    ContextType,
    FetcherCallableType,
    IdxResultType,
    ResolveType,
    json_dumps,
)

from ruamel.yaml.comments import CommentedMap, CommentedSeq

from . import CWL_CONTENT_TYPES, process, update
from .context import LoadingContext
from .errors import GraphTargetMissingException
from .loghandler import _logger
from .process import Process, get_schema, shortname
from .update import ALLUPDATES
from .utils import CWLObjectType, ResolverType, visit_class

jobloaderctx: ContextType = {
    "cwl": "https://w3id.org/cwl/cwl#",
    "cwltool": "http://commonwl.org/cwltool#",
    "path": {"@type": "@id"},
    "location": {"@type": "@id"},
    "id": "@id",
}


overrides_ctx: ContextType = {
    "overrideTarget": {"@type": "@id"},
    "cwltool": "http://commonwl.org/cwltool#",
    "http://commonwl.org/cwltool#overrides": {
        "@id": "cwltool:overrides",
        "mapSubject": "overrideTarget",
    },
    "requirements": {
        "@id": "https://w3id.org/cwl/cwl#requirements",
        "mapSubject": "class",
    },
}


def default_loader(
    fetcher_constructor: Optional[FetcherCallableType] = None,
    enable_dev: bool = False,
    doc_cache: bool = True,
) -> Loader:
    return Loader(
        jobloaderctx,
        fetcher_constructor=fetcher_constructor,
        allow_attachments=lambda r: enable_dev,
        doc_cache=doc_cache,
    )


def resolve_tool_uri(
    argsworkflow: str,
    resolver: Optional[ResolverType] = None,
    fetcher_constructor: Optional[FetcherCallableType] = None,
    document_loader: Optional[Loader] = None,
) -> Tuple[str, str]:
    uri = None  # type: Optional[str]
    split = urllib.parse.urlsplit(argsworkflow)
    # In case of Windows path, urlsplit misjudge Drive letters as scheme, here we are skipping that
    if split.scheme and split.scheme in ["http", "https", "file"]:
        uri = argsworkflow
    elif os.path.exists(os.path.abspath(argsworkflow)):
        uri = file_uri(str(os.path.abspath(argsworkflow)))
    elif resolver is not None:
        uri = resolver(document_loader or default_loader(fetcher_constructor), argsworkflow)

    if uri is None:
        raise ValidationException("Not found: '%s'" % argsworkflow)

    if argsworkflow != uri:
        _logger.info("Resolved '%s' to '%s'", argsworkflow, uri)

    fileuri = urllib.parse.urldefrag(uri)[0]
    return uri, fileuri


def fetch_document(
    argsworkflow: Union[str, CWLObjectType],
    loadingContext: Optional[LoadingContext] = None,
) -> Tuple[LoadingContext, CommentedMap, str]:
    """Retrieve a CWL document."""
    if loadingContext is None:
        loadingContext = LoadingContext()
        loadingContext.loader = default_loader()
    else:
        loadingContext = loadingContext.copy()
        if loadingContext.loader is None:
            loadingContext.loader = default_loader(
                loadingContext.fetcher_constructor,
                enable_dev=loadingContext.enable_dev,
                doc_cache=loadingContext.doc_cache,
            )

    if isinstance(argsworkflow, str):
        uri, fileuri = resolve_tool_uri(
            argsworkflow,
            resolver=loadingContext.resolver,
            document_loader=loadingContext.loader,
        )
        workflowobj = cast(
            CommentedMap,
            loadingContext.loader.fetch(fileuri, content_types=CWL_CONTENT_TYPES),
        )
        return loadingContext, workflowobj, uri
    if isinstance(argsworkflow, MutableMapping):
        uri = cast(str, argsworkflow["id"]) if argsworkflow.get("id") else "_:" + str(uuid.uuid4())
        workflowobj = cast(CommentedMap, cmap(cast(Dict[str, Any], argsworkflow), fn=uri))
        loadingContext.loader.idx[uri] = workflowobj
        return loadingContext, workflowobj, uri
    raise ValidationException("Must be URI or object: '%s'" % argsworkflow)


def _convert_stdstreams_to_files(
    workflowobj: Union[CWLObjectType, MutableSequence[Union[CWLObjectType, str, int]], str]
) -> None:
    if isinstance(workflowobj, MutableMapping):
        if workflowobj.get("class") == "CommandLineTool":
            with SourceLine(
                workflowobj,
                "outputs",
                ValidationException,
                _logger.isEnabledFor(logging.DEBUG),
            ):
                outputs = workflowobj.get("outputs", [])
                if not isinstance(outputs, CommentedSeq):
                    raise ValidationException('"outputs" section is not ' "valid.")
                for out in cast(MutableSequence[CWLObjectType], workflowobj.get("outputs", [])):
                    if not isinstance(out, CommentedMap):
                        raise ValidationException(f"Output {out!r} is not a valid OutputParameter.")
                    for streamtype in ["stdout", "stderr"]:
                        if out.get("type") == streamtype:
                            if "outputBinding" in out:
                                raise ValidationException(
                                    "Not allowed to specify outputBinding when"
                                    " using %s shortcut." % streamtype
                                )
                            if streamtype in workflowobj:
                                filename = workflowobj[streamtype]
                            else:
                                filename = str(
                                    hashlib.sha1(  # nosec
                                        json_dumps(workflowobj, sort_keys=True).encode("utf-8")
                                    ).hexdigest()
                                )
                                workflowobj[streamtype] = filename
                            out["type"] = "File"
                            out["outputBinding"] = cmap({"glob": filename})
            for inp in cast(MutableSequence[CWLObjectType], workflowobj.get("inputs", [])):
                if inp.get("type") == "stdin":
                    if "inputBinding" in inp:
                        raise ValidationException(
                            "Not allowed to specify inputBinding when" " using stdin shortcut."
                        )
                    if "stdin" in workflowobj:
                        raise ValidationException(
                            "Not allowed to specify stdin path when" " using stdin type shortcut."
                        )
                    else:
                        workflowobj["stdin"] = (
                            "$(inputs.%s.path)"
                            % cast(str, inp["id"]).rpartition("#")[2].split("/")[-1]
                        )
                        inp["type"] = "File"
        else:
            for entry in workflowobj.values():
                _convert_stdstreams_to_files(
                    cast(
                        Union[
                            CWLObjectType,
                            MutableSequence[Union[CWLObjectType, str, int]],
                            str,
                        ],
                        entry,
                    )
                )
    if isinstance(workflowobj, MutableSequence):
        for entry in workflowobj:
            _convert_stdstreams_to_files(
                cast(
                    Union[
                        CWLObjectType,
                        MutableSequence[Union[CWLObjectType, str, int]],
                        str,
                    ],
                    entry,
                )
            )


def _add_blank_ids(
    workflowobj: Union[CWLObjectType, MutableSequence[Union[CWLObjectType, str]]]
) -> None:
    if isinstance(workflowobj, MutableMapping):
        if (
            "run" in workflowobj
            and isinstance(workflowobj["run"], MutableMapping)
            and "id" not in workflowobj["run"]
            and "$import" not in workflowobj["run"]
        ):
            workflowobj["run"]["id"] = str(uuid.uuid4())
        for entry in workflowobj.values():
            _add_blank_ids(
                cast(
                    Union[CWLObjectType, MutableSequence[Union[CWLObjectType, str]]],
                    entry,
                )
            )
    if isinstance(workflowobj, MutableSequence):
        for entry in workflowobj:
            _add_blank_ids(
                cast(
                    Union[CWLObjectType, MutableSequence[Union[CWLObjectType, str]]],
                    entry,
                )
            )


def _fast_parser_convert_stdstreams_to_files(
    processobj: Union[cwl_v1_2.Process, MutableSequence[cwl_v1_2.Process]]
) -> None:
    if isinstance(processobj, cwl_v1_2.CommandLineTool):
        cwl_v1_2_utils.convert_stdstreams_to_files(processobj)
    elif isinstance(processobj, cwl_v1_2.Workflow):
        for st in processobj.steps:
            _fast_parser_convert_stdstreams_to_files(st.run)
    elif isinstance(processobj, MutableSequence):
        for p in processobj:
            _fast_parser_convert_stdstreams_to_files(p)


def _fast_parser_expand_hint_class(
    hints: Optional[Any], loadingOptions: cwl_v1_2.LoadingOptions
) -> None:
    if isinstance(hints, MutableSequence):
        for h in hints:
            if isinstance(h, MutableMapping) and "class" in h:
                for k, v in loadingOptions.namespaces.items():
                    if h["class"].startswith(k + ":"):
                        h["class"] = v + h["class"][len(k) + 1 :]


def _fast_parser_handle_hints(
    processobj: Union[cwl_v1_2.Process, MutableSequence[cwl_v1_2.Process]],
    loadingOptions: cwl_v1_2.LoadingOptions,
) -> None:
    if isinstance(processobj, (cwl_v1_2.CommandLineTool, cwl_v1_2.Workflow)):
        _fast_parser_expand_hint_class(processobj.hints, loadingOptions)

    if isinstance(processobj, cwl_v1_2.Workflow):
        for st in processobj.steps:
            _fast_parser_expand_hint_class(st.hints, loadingOptions)
            _fast_parser_handle_hints(st.run, loadingOptions)
    elif isinstance(processobj, MutableSequence):
        for p in processobj:
            _fast_parser_handle_hints(p, loadingOptions)


def update_index(document_loader: Loader, pr: CommentedMap) -> None:
    if "id" in pr:
        document_loader.idx[pr["id"]] = pr


def fast_parser(
    workflowobj: Union[CommentedMap, CommentedSeq, None],
    fileuri: Optional[str],
    uri: str,
    loadingContext: LoadingContext,
    fetcher: Fetcher,
) -> Tuple[Union[CommentedMap, CommentedSeq], CommentedMap]:
    lopt = cwl_v1_2.LoadingOptions(idx=loadingContext.codegen_idx, fileuri=fileuri, fetcher=fetcher)

    if uri not in loadingContext.codegen_idx:
        cwl_v1_2.load_document_with_metadata(
            workflowobj,
            fileuri,
            loadingOptions=lopt,
            addl_metadata_fields=["id", "cwlVersion"],
        )

    objects, loadopt = loadingContext.codegen_idx[uri]

    _fast_parser_convert_stdstreams_to_files(objects)
    _fast_parser_handle_hints(objects, loadopt)

    processobj: Union[MutableMapping[str, Any], MutableSequence[Any], float, str, None]

    processobj = cwl_v1_2.save(objects, relative_uris=False)

    metadata: Dict[str, Any] = {}
    metadata["id"] = loadopt.fileuri

    if loadopt.namespaces:
        metadata["$namespaces"] = loadopt.namespaces
    if loadopt.schemas:
        metadata["$schemas"] = loadopt.schemas
    if loadopt.baseuri:
        metadata["$base"] = loadopt.baseuri
    for k, v in loadopt.addl_metadata.items():
        if isinstance(processobj, MutableMapping) and k in processobj:
            metadata[k] = processobj[k]
        else:
            metadata[k] = v

    if loadingContext.loader:
        loadingContext.loader.graph += loadopt.graph

        # Need to match the document loader's index with the fast parser index
        # Get the base URI (no fragments) for documents that use $graph
        nofrag = urllib.parse.urldefrag(uri)[0]

        flag = "fastparser-idx-from:" + nofrag
        if not loadingContext.loader.idx.get(flag):
            objects, loadopt = loadingContext.codegen_idx[nofrag]
            fileobj = cmap(
                cast(
                    Union[int, float, str, Dict[str, Any], List[Any], None],
                    cwl_v1_2.save(objects, relative_uris=False),
                )
            )
            visit_class(
                fileobj,
                ("CommandLineTool", "Workflow", "ExpressionTool"),
                partial(update_index, loadingContext.loader),
            )
            loadingContext.loader.idx[flag] = flag
            for u in lopt.imports:
                loadingContext.loader.idx["import:" + u] = "import:" + u
            for u in lopt.includes:
                loadingContext.loader.idx["include:" + u] = "include:" + u

    return cast(
        Union[CommentedMap, CommentedSeq],
        cmap(cast(Union[Dict[str, Any], List[Any]], processobj)),
    ), cast(CommentedMap, cmap(metadata))


def resolve_and_validate_document(
    loadingContext: LoadingContext,
    workflowobj: Union[CommentedMap, CommentedSeq],
    uri: str,
    preprocess_only: bool = False,
) -> Tuple[LoadingContext, str]:
    """Validate a CWL document."""
    if not loadingContext.loader:
        raise ValueError("loadingContext must have a loader.")
    else:
        loader = loadingContext.loader
    loadingContext = loadingContext.copy()

    if not isinstance(workflowobj, MutableMapping):
        raise ValueError(
            "workflowjobj must be a dict, got '{}': {}".format(type(workflowobj), workflowobj)
        )

    jobobj = None
    if "cwl:tool" in workflowobj:
        jobobj, _ = loader.resolve_all(workflowobj, uri)
        uri = urllib.parse.urljoin(uri, workflowobj["https://w3id.org/cwl/cwl#tool"])
        del cast(Dict[str, Any], jobobj)["https://w3id.org/cwl/cwl#tool"]

        workflowobj = fetch_document(uri, loadingContext)[1]

    fileuri = urllib.parse.urldefrag(uri)[0]

    metadata: CWLObjectType

    cwlVersion = loadingContext.metadata.get("cwlVersion")
    if not cwlVersion:
        cwlVersion = workflowobj.get("cwlVersion")
    if not cwlVersion and fileuri != uri:
        # The tool we're loading is a fragment of a bigger file.  Get
        # the document root element and look for cwlVersion there.
        metadata = cast(CWLObjectType, fetch_document(fileuri, loadingContext)[1])
        cwlVersion = cast(str, metadata.get("cwlVersion"))
    if not cwlVersion:
        raise ValidationException(
            "No cwlVersion found. "
            "Use the following syntax in your CWL document to declare "
            "the version: cwlVersion: <version>.\n"
            "Note: if this is a CWL draft-3 (pre v1.0) document then it "
            "will need to be upgraded first using https://pypi.org/project/cwl-upgrader/ . "
            "'sbg:draft-2' documents can be upgraded using "
            "https://pypi.org/project/sevenbridges-cwl-draft2-upgrader/ ."
        )

    if not isinstance(cwlVersion, str):
        with SourceLine(workflowobj, "cwlVersion", ValidationException, loadingContext.debug):
            raise ValidationException(f"'cwlVersion' must be a string, got {type(cwlVersion)}")
    # strip out version
    cwlVersion = re.sub(r"^(?:cwl:|https://w3id.org/cwl/cwl#)", "", cwlVersion)
    if cwlVersion not in list(ALLUPDATES):
        # print out all the Supported Versions of cwlVersion
        versions = []
        for version in list(ALLUPDATES):
            if "dev" in version:
                version += " (with --enable-dev flag only)"
            versions.append(version)
        versions.sort()
        raise ValidationException(
            "The CWL reference runner no longer supports pre CWL v1.0 "
            "documents. Supported versions are: "
            "\n{}".format("\n".join(versions))
        )

    if isinstance(jobobj, CommentedMap) and "http://commonwl.org/cwltool#overrides" in jobobj:
        loadingContext.overrides_list.extend(resolve_overrides(jobobj, uri, uri))
        del jobobj["http://commonwl.org/cwltool#overrides"]

    if isinstance(jobobj, CommentedMap) and "https://w3id.org/cwl/cwl#requirements" in jobobj:
        if cwlVersion not in ("v1.1.0-dev1", "v1.1"):
            raise ValidationException(
                "`cwl:requirements` in the input object is not part of CWL "
                "v1.0. You can adjust to use `cwltool:overrides` instead; or you "
                "can set the cwlVersion to v1.1 or greater."
            )
        loadingContext.overrides_list.append(
            {
                "overrideTarget": uri,
                "requirements": jobobj["https://w3id.org/cwl/cwl#requirements"],
            }
        )
        del jobobj["https://w3id.org/cwl/cwl#requirements"]

    (sch_document_loader, avsc_names) = process.get_schema(cwlVersion)[:2]

    if isinstance(avsc_names, Exception):
        raise avsc_names

    processobj: ResolveType
    document_loader = Loader(
        sch_document_loader.ctx,
        schemagraph=sch_document_loader.graph,
        idx=loader.idx,
        cache=sch_document_loader.cache,
        fetcher_constructor=loadingContext.fetcher_constructor,
        skip_schemas=loadingContext.skip_schemas,
        doc_cache=loadingContext.doc_cache,
    )

    loadingContext.loader = document_loader

    if cwlVersion == "v1.0":
        _add_blank_ids(workflowobj)

    if cwlVersion != "v1.2":
        loadingContext.fast_parser = False

    if loadingContext.skip_resolve_all:
        # Some integrations (e.g. Arvados) loads documents, makes
        # in-memory changes to them (which are applied to the objects
        # in the document_loader index), and then sends them back
        # through the loading machinery.
        #
        # In this case, the functions of resolve_all() have already
        # happened.  Because resolve_all() is expensive, we don't want
        # to do it again if it's going to be a no-op, so the
        # skip_resolve_all flag tells us just to use the document
        # as-is from the loader index.
        #
        # Note that at the moment, fast_parser code path is considered
        # functionally the same as resolve_all() for this case.
        #
        processobj, metadata = document_loader.resolve_ref(uri)
    elif loadingContext.fast_parser:
        processobj, metadata = fast_parser(
            workflowobj, fileuri, uri, loadingContext, document_loader.fetcher
        )
    else:
        document_loader.resolve_all(workflowobj, fileuri)
        processobj, metadata = document_loader.resolve_ref(uri)

    if not isinstance(processobj, (CommentedMap, CommentedSeq)):
        raise ValidationException("Workflow must be a CommentedMap or CommentedSeq.")

    if not hasattr(processobj.lc, "filename"):
        processobj.lc.filename = fileuri

    if loadingContext.metadata:
        metadata = loadingContext.metadata

    # Make a shallow copy.  If we do a version update later, metadata
    # will be updated, we don't want to write through and change the
    # original object.
    metadata = copy.copy(metadata)

    if not isinstance(metadata, CommentedMap):
        raise ValidationException("metadata must be a CommentedMap, was %s" % type(metadata))

    if isinstance(processobj, CommentedMap):
        uri = processobj["id"]

    if not loadingContext.fast_parser:
        _convert_stdstreams_to_files(workflowobj)

    if isinstance(jobobj, CommentedMap):
        loadingContext.jobdefaults = jobobj

    loadingContext.avsc_names = avsc_names
    loadingContext.metadata = metadata

    if preprocess_only:
        return loadingContext, uri

    if loadingContext.do_validate:
        validate_doc(avsc_names, processobj, document_loader, loadingContext.strict)

    # None means default behavior (do update)
    if loadingContext.do_update in (True, None):
        if "cwlVersion" not in metadata:
            metadata["cwlVersion"] = cwlVersion
        processobj = update.update(
            processobj, document_loader, fileuri, loadingContext.enable_dev, metadata
        )
        document_loader.idx[processobj["id"]] = processobj

        visit_class(
            processobj,
            ("CommandLineTool", "Workflow", "ExpressionTool"),
            partial(update_index, document_loader),
        )

    return loadingContext, uri


def make_tool(
    uri: Union[str, CommentedMap, CommentedSeq], loadingContext: LoadingContext
) -> Process:
    """Make a Python CWL object."""
    if loadingContext.loader is None:
        raise ValueError("loadingContext must have a loader")

    resolveduri: Union[float, str, CommentedMap, CommentedSeq, None]
    metadata: CWLObjectType

    if loadingContext.fast_parser and isinstance(uri, str) and not loadingContext.skip_resolve_all:
        resolveduri, metadata = fast_parser(
            None, None, uri, loadingContext, loadingContext.loader.fetcher
        )
    else:
        resolveduri, metadata = loadingContext.loader.resolve_ref(uri)

    processobj = None
    if isinstance(resolveduri, MutableSequence):
        for obj in resolveduri:
            if obj["id"].endswith("#main"):
                processobj = obj
                break
        if not processobj:
            raise GraphTargetMissingException(
                "Tool file contains graph of multiple objects, must specify "
                "one of #%s"
                % ", #".join(urllib.parse.urldefrag(i["id"])[1] for i in resolveduri if "id" in i)
            )
    elif isinstance(resolveduri, MutableMapping):
        processobj = resolveduri
    else:
        raise Exception("Must resolve to list or dict")

    tool = loadingContext.construct_tool_object(processobj, loadingContext)

    if loadingContext.jobdefaults:
        jobobj = loadingContext.jobdefaults
        for inp in tool.tool["inputs"]:
            if shortname(inp["id"]) in jobobj:
                inp["default"] = jobobj[shortname(inp["id"])]

    return tool


def load_tool(
    argsworkflow: Union[str, CWLObjectType],
    loadingContext: Optional[LoadingContext] = None,
) -> Process:
    loadingContext, workflowobj, uri = fetch_document(argsworkflow, loadingContext)

    loadingContext, uri = resolve_and_validate_document(
        loadingContext,
        workflowobj,
        uri,
    )

    return make_tool(uri, loadingContext)


def resolve_overrides(
    ov: IdxResultType,
    ov_uri: str,
    baseurl: str,
) -> List[CWLObjectType]:
    ovloader = Loader(overrides_ctx)
    ret, _ = ovloader.resolve_all(ov, baseurl)
    if not isinstance(ret, CommentedMap):
        raise Exception("Expected CommentedMap, got %s" % type(ret))
    cwl_docloader = get_schema("v1.0")[0]
    cwl_docloader.resolve_all(ret, ov_uri)
    return cast(List[CWLObjectType], ret["http://commonwl.org/cwltool#overrides"])


def load_overrides(ov: str, base_url: str) -> List[CWLObjectType]:
    ovloader = Loader(overrides_ctx)
    return resolve_overrides(ovloader.fetch(ov), ov, base_url)


def recursive_resolve_and_validate_document(
    loadingContext: LoadingContext,
    workflowobj: Union[CommentedMap, CommentedSeq],
    uri: str,
    preprocess_only: bool = False,
) -> Tuple[LoadingContext, str, Process]:
    """Validate a CWL document, checking that a tool object can be built."""
    loadingContext, uri = resolve_and_validate_document(
        loadingContext,
        workflowobj,
        uri,
        preprocess_only=preprocess_only,
    )
    tool = make_tool(uri, loadingContext)
    return loadingContext, uri, tool



# Stubs for networkx.drawing.layout (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def random_layout(
    G, center: Optional[Any] = ..., dim: int = ..., random_state: Optional[Any] = ...
): ...
def circular_layout(G, scale: int = ..., center: Optional[Any] = ..., dim: int = ...): ...
def shell_layout(
    G,
    nlist: Optional[Any] = ...,
    scale: int = ...,
    center: Optional[Any] = ...,
    dim: int = ...,
): ...
def fruchterman_reingold_layout(
    G,
    k: Optional[Any] = ...,
    pos: Optional[Any] = ...,
    fixed: Optional[Any] = ...,
    iterations: int = ...,
    threshold: float = ...,
    weight: str = ...,
    scale: int = ...,
    center: Optional[Any] = ...,
    dim: int = ...,
    random_state: Optional[Any] = ...,
): ...

spring_layout = fruchterman_reingold_layout

def kamada_kawai_layout(
    G,
    dist: Optional[Any] = ...,
    pos: Optional[Any] = ...,
    weight: str = ...,
    scale: int = ...,
    center: Optional[Any] = ...,
    dim: int = ...,
): ...
def spectral_layout(
    G, weight: str = ..., scale: int = ..., center: Optional[Any] = ..., dim: int = ...
): ...
def rescale_layout(pos, scale: int = ...): ...



# Stubs for networkx.readwrite.json_graph.node_link (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def node_link_data(G, attrs: Optional[Any] = ...): ...
def node_link_graph(
    data, directed: bool = ..., multigraph: bool = ..., attrs: Optional[Any] = ...
): ...



# Stubs for networkx.algorithms.hybrid (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def kl_connected_subgraph(G, k, l, low_memory: bool = ..., same_as_graph: bool = ...): ...
def is_kl_connected(G, k, l, low_memory: bool = ...): ...



# Stubs for networkx.algorithms.isomorphism.matchhelpers (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

def categorical_node_match(attr, default): ...

categorical_edge_match: Any

def categorical_multiedge_match(attr, default): ...
def numerical_node_match(attr, default, rtol: float = ..., atol: float = ...): ...

numerical_edge_match: Any

def numerical_multiedge_match(attr, default, rtol: float = ..., atol: float = ...): ...
def generic_node_match(attr, default, op): ...

generic_edge_match: Any

def generic_multiedge_match(attr, default, op): ...



"""Enables Docker software containers via the udocker runtime."""

from typing import List

from .docker import DockerCommandLineJob


class UDockerCommandLineJob(DockerCommandLineJob):
    """Runs a CommandLineJob in a software container using the udocker engine."""

    @staticmethod
    def append_volume(runtime: List[str], source: str, target: str, writable: bool = False) -> None:
        """Add binding arguments to the runtime list."""
        runtime.append("--volume={}:{}:{}".format(source, target, "rw" if writable else "ro"))



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class SH(DefinedNamespace):
    alternativePath: URIRef
    annotationProperty: URIRef
    annotationValue: URIRef
    annotationVarName: URIRef
    ask: URIRef
    closed: URIRef
    condition: URIRef
    conforms: URIRef
    construct: URIRef
    datatype: URIRef
    deactivated: URIRef
    declare: URIRef
    defaultValue: URIRef
    description: URIRef
    detail: URIRef
    disjoint: URIRef
    entailment: URIRef
    equals: URIRef
    expression: URIRef
    filterShape: URIRef
    flags: URIRef
    focusNode: URIRef
    group: URIRef
    hasValue: URIRef
    ignoredProperties: URIRef
    intersection: URIRef
    inversePath: URIRef
    js: URIRef
    jsFunctionName: URIRef
    jsLibrary: URIRef
    jsLibraryURL: URIRef
    labelTemplate: URIRef
    languageIn: URIRef
    lessThan: URIRef
    lessThanOrEquals: URIRef
    maxCount: URIRef
    maxExclusive: URIRef
    maxInclusive: URIRef
    maxLength: URIRef
    message: URIRef
    minCount: URIRef
    minExclusive: URIRef
    minInclusive: URIRef
    minLength: URIRef
    name: URIRef
    namespace: URIRef
    node: URIRef
    nodeKind: URIRef
    nodeValidator: URIRef
    nodes: URIRef
    object: URIRef
    oneOrMorePath: URIRef
    optional: URIRef
    order: URIRef
    parameter: URIRef
    path: URIRef
    pattern: URIRef
    predicate: URIRef
    prefix: URIRef
    prefixes: URIRef
    property: URIRef
    propertyValidator: URIRef
    qualifiedMaxCount: URIRef
    qualifiedMinCount: URIRef
    qualifiedValueShape: URIRef
    qualifiedValueShapesDisjoint: URIRef
    result: URIRef
    resultAnnotation: URIRef
    resultMessage: URIRef
    resultPath: URIRef
    resultSeverity: URIRef
    returnType: URIRef
    rule: URIRef
    select: URIRef
    severity: URIRef
    shapesGraph: URIRef
    shapesGraphWellFormed: URIRef
    sourceConstraint: URIRef
    sourceConstraintComponent: URIRef
    sourceShape: URIRef
    sparql: URIRef
    subject: URIRef
    suggestedShapesGraph: URIRef
    target: URIRef
    targetClass: URIRef
    targetNode: URIRef
    targetObjectsOf: URIRef
    targetSubjectsOf: URIRef
    union: URIRef
    uniqueLang: URIRef
    update: URIRef
    validator: URIRef
    value: URIRef
    xone: URIRef
    zeroOrMorePath: URIRef
    zeroOrOnePath: URIRef
    AbstractResult: URIRef
    ConstraintComponent: URIRef
    Function: URIRef
    JSConstraint: URIRef
    JSExecutable: URIRef
    JSFunction: URIRef
    JSLibrary: URIRef
    JSRule: URIRef
    JSTarget: URIRef
    JSTargetType: URIRef
    JSValidator: URIRef
    NodeKind: URIRef
    NodeShape: URIRef
    Parameter: URIRef
    Parameterizable: URIRef
    PrefixDeclaration: URIRef
    PropertyGroup: URIRef
    PropertyShape: URIRef
    ResultAnnotation: URIRef
    Rule: URIRef
    SPARQLAskExecutable: URIRef
    SPARQLAskValidator: URIRef
    SPARQLConstraint: URIRef
    SPARQLConstructExecutable: URIRef
    SPARQLExecutable: URIRef
    SPARQLFunction: URIRef
    SPARQLRule: URIRef
    SPARQLSelectExecutable: URIRef
    SPARQLSelectValidator: URIRef
    SPARQLTarget: URIRef
    SPARQLTargetType: URIRef
    SPARQLUpdateExecutable: URIRef
    Severity: URIRef
    Shape: URIRef
    Target: URIRef
    TargetType: URIRef
    TripleRule: URIRef
    ValidationReport: URIRef
    ValidationResult: URIRef
    Validator: URIRef
    this: URIRef
    AndConstraintComponent: URIRef
    ClassConstraintComponent: URIRef
    ClosedConstraintComponent: URIRef
    DatatypeConstraintComponent: URIRef
    DisjointConstraintComponent: URIRef
    EqualsConstraintComponent: URIRef
    ExpressionConstraintComponent: URIRef
    HasValueConstraintComponent: URIRef
    InConstraintComponent: URIRef
    JSConstraintComponent: URIRef
    LanguageInConstraintComponent: URIRef
    LessThanConstraintComponent: URIRef
    LessThanOrEqualsConstraintComponent: URIRef
    MaxCountConstraintComponent: URIRef
    MaxExclusiveConstraintComponent: URIRef
    MaxInclusiveConstraintComponent: URIRef
    MaxLengthConstraintComponent: URIRef
    MinCountConstraintComponent: URIRef
    MinExclusiveConstraintComponent: URIRef
    MinInclusiveConstraintComponent: URIRef
    MinLengthConstraintComponent: URIRef
    NodeConstraintComponent: URIRef
    NodeKindConstraintComponent: URIRef
    NotConstraintComponent: URIRef
    OrConstraintComponent: URIRef
    PatternConstraintComponent: URIRef
    PropertyConstraintComponent: URIRef
    QualifiedMaxCountConstraintComponent: URIRef
    QualifiedMinCountConstraintComponent: URIRef
    SPARQLConstraintComponent: URIRef
    UniqueLangConstraintComponent: URIRef
    XoneConstraintComponent: URIRef
    BlankNode: URIRef
    BlankNodeOrIRI: URIRef
    BlankNodeOrLiteral: URIRef
    IRI: URIRef
    IRIOrLiteral: URIRef
    Literal: URIRef
    Info: URIRef
    Violation: URIRef
    Warning: URIRef



# Stubs for networkx.algorithms.bipartite.edgelist (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def write_edgelist(
    G,
    path,
    comments: str = ...,
    delimiter: str = ...,
    data: bool = ...,
    encoding: str = ...,
): ...
def generate_edgelist(G, delimiter: str = ..., data: bool = ...): ...
def parse_edgelist(
    lines,
    comments: str = ...,
    delimiter: Optional[Any] = ...,
    create_using: Optional[Any] = ...,
    nodetype: Optional[Any] = ...,
    data: bool = ...,
): ...
def read_edgelist(
    path,
    comments: str = ...,
    delimiter: Optional[Any] = ...,
    create_using: Optional[Any] = ...,
    nodetype: Optional[Any] = ...,
    data: bool = ...,
    edgetype: Optional[Any] = ...,
    encoding: str = ...,
): ...



# Stubs for networkx.algorithms.centrality.harmonic (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def harmonic_centrality(G, nbunch: Optional[Any] = ..., distance: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.tree.mst (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def minimum_spanning_edges(
    G,
    algorithm: str = ...,
    weight: str = ...,
    keys: bool = ...,
    data: bool = ...,
    ignore_nan: bool = ...,
): ...
def maximum_spanning_edges(
    G,
    algorithm: str = ...,
    weight: str = ...,
    keys: bool = ...,
    data: bool = ...,
    ignore_nan: bool = ...,
): ...
def minimum_spanning_tree(G, weight: str = ..., algorithm: str = ..., ignore_nan: bool = ...): ...
def maximum_spanning_tree(G, weight: str = ..., algorithm: str = ..., ignore_nan: bool = ...): ...



import json
from pathlib import Path

from .util import get_data, get_main_output


def test_log_dir_echo_output(tmp_path: Path) -> None:
    _, stdout, stderr = get_main_output(
        ["--log-dir", str(tmp_path), get_data("tests/echo.cwl"), "--inp", "hello"]
    )
    assert "completed success" in stderr, stderr
    assert json.loads(stdout)["out"].strip("\n") == "hello"
    assert len(list(tmp_path.iterdir())) == 1
    subdir = next(tmp_path.iterdir())
    assert subdir.is_dir()
    assert len(list(subdir.iterdir())) == 1
    result = next(subdir.iterdir())
    assert result.name == "out.txt"
    output = open(result).read()
    assert output == "hello\n"


def test_log_dir_echo_no_output(tmp_path: Path) -> None:
    _, stdout, stderr = get_main_output(
        [
            "--log-dir",
            str(tmp_path),
            get_data("tests/echo-stdout-log-dir.cwl"),
            "--inp",
            "hello",
        ]
    )
    assert len(list(tmp_path.iterdir())) == 1
    subdir = next(tmp_path.iterdir())
    assert subdir.name == "echo"
    assert subdir.is_dir()
    assert len(list(subdir.iterdir())) == 1
    result = next(subdir.iterdir())
    assert result.name == "out.txt"
    output = open(result).read()
    assert output == "hello\n"


def test_log_dir_echo_stderr(tmp_path: Path) -> None:
    _, stdout, stderr = get_main_output(
        [
            "--log-dir",
            str(tmp_path),
            get_data("tests/echo-stderr.cwl"),
            "--message",
            "hello",
        ]
    )
    assert len(list(tmp_path.iterdir())) == 1
    subdir = next(tmp_path.iterdir())
    assert subdir.name == "echo-stderr.cwl"
    assert subdir.is_dir()
    assert len(list(subdir.iterdir())) == 1
    result = next(subdir.iterdir())
    assert result.name == "out.txt"
    output = open(result).read()
    assert output == "hello\n"



# Stubs for networkx.classes.graphviews (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

from networkx.classes import DiGraph, Graph, MultiDiGraph, MultiGraph
from networkx.classes.coreviews import ReadOnlyGraph

class SubGraph(ReadOnlyGraph, Graph):
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, graph, filter_node: Any = ..., filter_edge: Any = ...) -> None: ...

class SubDiGraph(ReadOnlyGraph, DiGraph):
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, graph, filter_node: Any = ..., filter_edge: Any = ...) -> None: ...

class SubMultiGraph(ReadOnlyGraph, MultiGraph):
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, graph, filter_node: Any = ..., filter_edge: Any = ...) -> None: ...

class SubMultiDiGraph(ReadOnlyGraph, MultiDiGraph):
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, graph, filter_node: Any = ..., filter_edge: Any = ...) -> None: ...

class ReverseView(ReadOnlyGraph, DiGraph):
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, graph) -> None: ...

class MultiReverseView(ReadOnlyGraph, MultiDiGraph):
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, graph) -> None: ...

class DiGraphView(ReadOnlyGraph, DiGraph):
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, graph) -> None: ...

class MultiDiGraphView(ReadOnlyGraph, MultiDiGraph):
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, graph) -> None: ...

class GraphView(ReadOnlyGraph, Graph):
    UnionAdj: Any = ...
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, graph) -> None: ...

class MultiGraphView(ReadOnlyGraph, MultiGraph):
    UnionAdj: Any = ...
    root_graph: Any = ...
    graph: Any = ...
    def __init__(self, graph) -> None: ...



# Stubs for networkx.algorithms.flow.gomory_hu (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from .edmondskarp import edmonds_karp

default_flow_func = edmonds_karp

def gomory_hu_tree(G, capacity: str = ..., flow_func: Optional[Any] = ...): ...



# Stubs for graphviz.backend (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

class ExecutableNotFound(RuntimeError):
    def __init__(self, args) -> None: ...

def render(engine, format, filepath, quiet: bool = ...): ...
def pipe(engine, format, data, quiet: bool = ...): ...
def version(): ...
def view(filepath): ...



# Stubs for networkx.algorithms.matching (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def maximal_matching(G): ...
def is_matching(G, matching): ...
def is_maximal_matching(G, matching): ...
def max_weight_matching(G, maxcardinality: bool = ..., weight: str = ...): ...



# Stubs for networkx.algorithms.clique (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def enumerate_all_cliques(G): ...
def find_cliques(G): ...
def find_cliques_recursive(G): ...
def make_max_clique_graph(G, create_using: Optional[Any] = ...): ...
def make_clique_bipartite(
    G,
    fpos: Optional[Any] = ...,
    create_using: Optional[Any] = ...,
    name: Optional[Any] = ...,
): ...
def graph_clique_number(G, cliques: Optional[Any] = ...): ...
def graph_number_of_cliques(G, cliques: Optional[Any] = ...): ...
def node_clique_number(G, nodes: Optional[Any] = ..., cliques: Optional[Any] = ...): ...
def number_of_cliques(G, nodes: Optional[Any] = ..., cliques: Optional[Any] = ...): ...
def cliques_containing_node(G, nodes: Optional[Any] = ..., cliques: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.bipartite.projection (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def projected_graph(B, nodes, multigraph: bool = ...): ...
def weighted_projected_graph(B, nodes, ratio: bool = ...): ...
def collaboration_weighted_projected_graph(B, nodes): ...
def overlap_weighted_projected_graph(B, nodes, jaccard: bool = ...): ...
def generic_weighted_projected_graph(B, nodes, weight_function: Optional[Any] = ...): ...
def project(B, nodes, create_using: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.coloring (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.algorithms.coloring.greedy_coloring import *

# Names in __all__ with no definition:
#   greedy_color



"""Test the recursive validation feature (validate document and try to build tool)."""
from cwltool.load_tool import fetch_document, recursive_resolve_and_validate_document

from .util import get_data


def test_recursive_validation() -> None:
    """Test the recursive_resolve_and_validate_document function."""
    loadingContext, workflowobj, uri = fetch_document(get_data("tests/wf/default_path.cwl"))
    loadingContext, uri, tool = recursive_resolve_and_validate_document(
        loadingContext, workflowobj, uri
    )



# Stubs for networkx.convert (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def to_networkx_graph(data, create_using: Optional[Any] = ..., multigraph_input: bool = ...): ...
def to_dict_of_lists(G, nodelist: Optional[Any] = ...): ...
def from_dict_of_lists(d, create_using: Optional[Any] = ...): ...
def to_dict_of_dicts(G, nodelist: Optional[Any] = ..., edge_data: Optional[Any] = ...): ...
def from_dict_of_dicts(d, create_using: Optional[Any] = ..., multigraph_input: bool = ...): ...
def to_edgelist(G, nodelist: Optional[Any] = ...): ...
def from_edgelist(edgelist, create_using: Optional[Any] = ...): ...



import asyncio
from concurrent.futures import Executor
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Iterator,
    List,
    MutableMapping,
    Optional,
    Pattern,
    Set,
    Sized,
    Tuple,
    Union,
)

from black.mode import Mode as Mode
from black.mode import TargetVersion as TargetVersion

FileContent = str
Encoding = str
NewLine = str
FileMode = Mode

def format_str(src_contents: str, mode: Mode) -> FileContent: ...



# Stubs for networkx.algorithms.flow (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from .boykovkolmogorov import *
from .capacityscaling import *
from .dinitz_alg import *
from .edmondskarp import *
from .gomory_hu import *
from .maxflow import *
from .mincost import *
from .networksimplex import *
from .preflowpush import *
from .shortestaugmentingpath import *
from .utils import build_flow_dict, build_residual_network



# Stubs for networkx.algorithms.chains (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def chain_decomposition(G, root: Optional[Any] = ...): ...



# Stubs for networkx.linalg.algebraicconnectivity (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

class _PCGSolver:
    def __init__(self, A, M) -> None: ...
    def solve(self, B, tol): ...

class _CholeskySolver:
    def __init__(self, A) -> None: ...
    def solve(self, B, tol: Optional[Any] = ...): ...

class _LUSolver:
    def __init__(self, A) -> None: ...
    def solve(self, B, tol: Optional[Any] = ...): ...

def algebraic_connectivity(
    G, weight: str = ..., normalized: bool = ..., tol: float = ..., method: str = ...
): ...
def fiedler_vector(
    G, weight: str = ..., normalized: bool = ..., tol: float = ..., method: str = ...
): ...
def spectral_ordering(
    G, weight: str = ..., normalized: bool = ..., tol: float = ..., method: str = ...
): ...
def setup_module(module): ...



# Stubs for networkx.convert_matrix (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def to_pandas_adjacency(
    G,
    nodelist: Optional[Any] = ...,
    dtype: Optional[Any] = ...,
    order: Optional[Any] = ...,
    multigraph_weight: Any = ...,
    weight: str = ...,
    nonedge: float = ...,
): ...
def from_pandas_adjacency(df, create_using: Optional[Any] = ...): ...
def to_pandas_edgelist(
    G,
    source: str = ...,
    target: str = ...,
    nodelist: Optional[Any] = ...,
    dtype: Optional[Any] = ...,
    order: Optional[Any] = ...,
): ...
def from_pandas_edgelist(
    df,
    source: str = ...,
    target: str = ...,
    edge_attr: Optional[Any] = ...,
    create_using: Optional[Any] = ...,
): ...
def to_numpy_matrix(
    G,
    nodelist: Optional[Any] = ...,
    dtype: Optional[Any] = ...,
    order: Optional[Any] = ...,
    multigraph_weight: Any = ...,
    weight: str = ...,
    nonedge: float = ...,
): ...
def from_numpy_matrix(A, parallel_edges: bool = ..., create_using: Optional[Any] = ...): ...
def to_numpy_recarray(
    G,
    nodelist: Optional[Any] = ...,
    dtype: Optional[Any] = ...,
    order: Optional[Any] = ...,
): ...
def to_scipy_sparse_matrix(
    G,
    nodelist: Optional[Any] = ...,
    dtype: Optional[Any] = ...,
    weight: str = ...,
    format: str = ...,
): ...
def from_scipy_sparse_matrix(
    A,
    parallel_edges: bool = ...,
    create_using: Optional[Any] = ...,
    edge_attribute: str = ...,
): ...
def to_numpy_array(
    G,
    nodelist: Optional[Any] = ...,
    dtype: Optional[Any] = ...,
    order: Optional[Any] = ...,
    multigraph_weight: Any = ...,
    weight: str = ...,
    nonedge: float = ...,
): ...
def from_numpy_array(A, parallel_edges: bool = ..., create_using: Optional[Any] = ...): ...



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class DCTERMS(DefinedNamespace):
    DCMIType: URIRef
    DDC: URIRef
    IMT: URIRef
    LCC: URIRef
    LCSH: URIRef
    MESH: URIRef
    NLM: URIRef
    TGN: URIRef
    UDC: URIRef
    abstract: URIRef
    accessRights: URIRef
    accrualMethod: URIRef
    accrualPeriodicity: URIRef
    accrualPolicy: URIRef
    alternative: URIRef
    audience: URIRef
    available: URIRef
    bibliographicCitation: URIRef
    conformsTo: URIRef
    contributor: URIRef
    coverage: URIRef
    created: URIRef
    creator: URIRef
    date: URIRef
    dateAccepted: URIRef
    dateCopyrighted: URIRef
    dateSubmitted: URIRef
    description: URIRef
    educationLevel: URIRef
    extent: URIRef
    format: URIRef
    hasFormat: URIRef
    hasPart: URIRef
    hasVersion: URIRef
    identifier: URIRef
    instructionalMethod: URIRef
    isFormatOf: URIRef
    isPartOf: URIRef
    isReferencedBy: URIRef
    isReplacedBy: URIRef
    isRequiredBy: URIRef
    isVersionOf: URIRef
    issued: URIRef
    language: URIRef
    license: URIRef
    mediator: URIRef
    medium: URIRef
    modified: URIRef
    provenance: URIRef
    publisher: URIRef
    references: URIRef
    relation: URIRef
    replaces: URIRef
    requires: URIRef
    rights: URIRef
    rightsHolder: URIRef
    source: URIRef
    spatial: URIRef
    subject: URIRef
    tableOfContents: URIRef
    temporal: URIRef
    title: URIRef
    type: URIRef
    valid: URIRef
    Agent: URIRef
    AgentClass: URIRef
    BibliographicResource: URIRef
    FileFormat: URIRef
    Frequency: URIRef
    Jurisdiction: URIRef
    LicenseDocument: URIRef
    LinguisticSystem: URIRef
    Location: URIRef
    LocationPeriodOrJurisdiction: URIRef
    MediaType: URIRef
    MediaTypeOrExtent: URIRef
    MethodOfAccrual: URIRef
    MethodOfInstruction: URIRef
    PeriodOfTime: URIRef
    PhysicalMedium: URIRef
    PhysicalResource: URIRef
    Policy: URIRef
    ProvenanceStatement: URIRef
    RightsStatement: URIRef
    SizeOrDuration: URIRef
    Standard: URIRef
    Box: URIRef
    ISO3166: URIRef
    Period: URIRef
    Point: URIRef
    RFC1766: URIRef
    RFC3066: URIRef
    RFC4646: URIRef
    RFC5646: URIRef
    URI: URIRef
    W3CDTF: URIRef



# Stubs for networkx.generators.community (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def caveman_graph(l, k): ...
def connected_caveman_graph(l, k): ...
def relaxed_caveman_graph(l, k, p, seed: Optional[Any] = ...): ...
def random_partition_graph(sizes, p_in, p_out, seed: Optional[Any] = ..., directed: bool = ...): ...
def planted_partition_graph(l, k, p_in, p_out, seed: Optional[Any] = ..., directed: bool = ...): ...
def gaussian_random_partition_graph(
    n, s, v, p_in, p_out, directed: bool = ..., seed: Optional[Any] = ...
): ...
def ring_of_cliques(num_cliques, clique_size): ...
def windmill_graph(n, k): ...



# Stubs for networkx.generators.degree_seq (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def configuration_model(
    deg_sequence, create_using: Optional[Any] = ..., seed: Optional[Any] = ...
): ...
def directed_configuration_model(
    in_degree_sequence,
    out_degree_sequence,
    create_using: Optional[Any] = ...,
    seed: Optional[Any] = ...,
): ...
def expected_degree_graph(w, seed: Optional[Any] = ..., selfloops: bool = ...): ...
def havel_hakimi_graph(deg_sequence, create_using: Optional[Any] = ...): ...
def directed_havel_hakimi_graph(
    in_deg_sequence, out_deg_sequence, create_using: Optional[Any] = ...
): ...
def degree_sequence_tree(deg_sequence, create_using: Optional[Any] = ...): ...
def random_degree_sequence_graph(sequence, seed: Optional[Any] = ..., tries: int = ...): ...

class DegreeSequenceRandomGraph:
    degree: Any = ...
    m: Any = ...
    dmax: Any = ...
    def __init__(self, degree, seed: Optional[Any] = ...) -> None: ...
    remaining_degree: Any = ...
    graph: Any = ...
    def generate(self): ...
    def update_remaining(self, u, v, aux_graph: Optional[Any] = ...): ...
    def p(self, u, v): ...
    def q(self, u, v): ...
    def suitable_edge(self): ...
    def phase1(self): ...
    def phase2(self): ...
    def phase3(self): ...



# Stubs for networkx.algorithms.centrality.flow_matrix (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def flow_matrix_row(G, weight: Optional[Any] = ..., dtype: Any = ..., solver: str = ...): ...

class InverseLaplacian:
    dtype: Any = ...
    n: Any = ...
    w: Any = ...
    C: Any = ...
    L1: Any = ...
    def __init__(self, L, width: Optional[Any] = ..., dtype: Optional[Any] = ...) -> None: ...
    def init_solver(self, L): ...
    def solve(self, r): ...
    def solve_inverse(self, r): ...
    def get_rows(self, r1, r2): ...
    def get_row(self, r): ...
    def width(self, L): ...

class FullInverseLaplacian(InverseLaplacian):
    IL: Any = ...
    def init_solver(self, L): ...
    def solve(self, rhs): ...
    def solve_inverse(self, r): ...

class SuperLUInverseLaplacian(InverseLaplacian):
    lusolve: Any = ...
    def init_solver(self, L): ...
    def solve_inverse(self, r): ...
    def solve(self, rhs): ...

class CGInverseLaplacian(InverseLaplacian):
    M: Any = ...
    def init_solver(self, L): ...
    def solve(self, rhs): ...
    def solve_inverse(self, r): ...

def laplacian_sparse_matrix(
    G,
    nodelist: Optional[Any] = ...,
    weight: Optional[Any] = ...,
    dtype: Optional[Any] = ...,
    format: str = ...,
): ...



# Stubs for networkx.algorithms (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.algorithms.assortativity import *
from networkx.algorithms.boundary import *
from networkx.algorithms.bridges import *
from networkx.algorithms.centrality import *
from networkx.algorithms.chains import *
from networkx.algorithms.chordal import *
from networkx.algorithms.clique import *
from networkx.algorithms.cluster import *
from networkx.algorithms.coloring import *
from networkx.algorithms.communicability_alg import *
from networkx.algorithms.components import *
from networkx.algorithms.core import *
from networkx.algorithms.covering import *
from networkx.algorithms.cuts import *
from networkx.algorithms.cycles import *
from networkx.algorithms.dag import *
from networkx.algorithms.distance_measures import *
from networkx.algorithms.distance_regular import *
from networkx.algorithms.dominance import *
from networkx.algorithms.dominating import *
from networkx.algorithms.efficiency import *
from networkx.algorithms.euler import *
from networkx.algorithms.graphical import *
from networkx.algorithms.hierarchy import *
from networkx.algorithms.hybrid import *
from networkx.algorithms.isolate import *
from networkx.algorithms.link_analysis import *
from networkx.algorithms.link_prediction import *
from networkx.algorithms.lowest_common_ancestors import *
from networkx.algorithms.matching import *
from networkx.algorithms.minors import *
from networkx.algorithms.mis import *
from networkx.algorithms.operators import *
from networkx.algorithms.reciprocity import *
from networkx.algorithms.richclub import *
from networkx.algorithms.shortest_paths import *
from networkx.algorithms.similarity import *
from networkx.algorithms.simple_paths import *
from networkx.algorithms.smetric import *
from networkx.algorithms.structuralholes import *
from networkx.algorithms.swap import *
from networkx.algorithms.traversal import *
from networkx.algorithms.tree.coding import *
from networkx.algorithms.tree.mst import *
from networkx.algorithms.tree.operations import *
from networkx.algorithms.tree.recognition import *
from networkx.algorithms.triads import *
from networkx.algorithms.vitality import *
from networkx.algorithms.voronoi import *
from networkx.algorithms.wiener import *



# Stubs for networkx.readwrite.json_graph.tree (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

def tree_data(G, root, attrs: Any = ...): ...
def tree_graph(data, attrs: Any = ...): ...



# Stubs for networkx.linalg.attrmatrix (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def attr_matrix(
    G,
    edge_attr: Optional[Any] = ...,
    node_attr: Optional[Any] = ...,
    normalized: bool = ...,
    rc_order: Optional[Any] = ...,
    dtype: Optional[Any] = ...,
    order: Optional[Any] = ...,
): ...
def attr_sparse_matrix(
    G,
    edge_attr: Optional[Any] = ...,
    node_attr: Optional[Any] = ...,
    normalized: bool = ...,
    rc_order: Optional[Any] = ...,
    dtype: Optional[Any] = ...,
): ...



# Stubs for networkx.utils (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.utils.contextmanagers import *
from networkx.utils.decorators import *
from networkx.utils.heaps import *
from networkx.utils.misc import *
from networkx.utils.random_sequence import *
from networkx.utils.rcm import *
from networkx.utils.union_find import *



# Stubs for networkx.algorithms.structuralholes (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def effective_size(G, nodes: Optional[Any] = ..., weight: Optional[Any] = ...): ...
def constraint(G, nodes: Optional[Any] = ..., weight: Optional[Any] = ...): ...
def local_constraint(G, u, v, weight: Optional[Any] = ...): ...



"""ProcessGenerator related tests."""

from pathlib import Path

import pytest

from cwltool.main import main

from .util import get_data


def test_missing_enable_ext(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    """Test missing enable-ext option fails.

    Check that a workflow that needs `--enable-ext` and
    `--enable-dev` fails without those options and passes with them.
    """
    monkeypatch.delenv("CWLTOOL_OPTIONS", raising=False)
    assert (
        main(
            [
                "--outdir",
                str(tmp_path),
                get_data("tests/wf/generator/zing.cwl"),
                "--zing",
                "zipper",
            ]
        )
        == 1
    )

    assert (
        main(
            [
                "--debug",
                "--enable-ext",
                "--enable-dev",
                "--outdir",
                str(tmp_path),
                get_data("tests/wf/generator/zing.cwl"),
                "--zing",
                "zipper",
            ]
        )
        == 0
    )

    monkeypatch.setenv("CWLTOOL_OPTIONS", "--enable-ext --enable-dev")
    assert (
        main(
            [
                "--outdir",
                str(tmp_path),
                get_data("tests/wf/generator/zing.cwl"),
                "--zing",
                "zipper",
            ]
        )
        == 0
    )



# Stubs for networkx.generators.random_clustered (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def random_clustered_graph(
    joint_degree_sequence, create_using: Optional[Any] = ..., seed: Optional[Any] = ...
): ...



# Stubs for networkx.generators.small (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def make_small_graph(graph_description, create_using: Optional[Any] = ...): ...
def LCF_graph(n, shift_list, repeats, create_using: Optional[Any] = ...): ...
def bull_graph(create_using: Optional[Any] = ...): ...
def chvatal_graph(create_using: Optional[Any] = ...): ...
def cubical_graph(create_using: Optional[Any] = ...): ...
def desargues_graph(create_using: Optional[Any] = ...): ...
def diamond_graph(create_using: Optional[Any] = ...): ...
def dodecahedral_graph(create_using: Optional[Any] = ...): ...
def frucht_graph(create_using: Optional[Any] = ...): ...
def heawood_graph(create_using: Optional[Any] = ...): ...
def hoffman_singleton_graph(): ...
def house_graph(create_using: Optional[Any] = ...): ...
def house_x_graph(create_using: Optional[Any] = ...): ...
def icosahedral_graph(create_using: Optional[Any] = ...): ...
def krackhardt_kite_graph(create_using: Optional[Any] = ...): ...
def moebius_kantor_graph(create_using: Optional[Any] = ...): ...
def octahedral_graph(create_using: Optional[Any] = ...): ...
def pappus_graph(): ...
def petersen_graph(create_using: Optional[Any] = ...): ...
def sedgewick_maze_graph(create_using: Optional[Any] = ...): ...
def tetrahedral_graph(create_using: Optional[Any] = ...): ...
def truncated_cube_graph(create_using: Optional[Any] = ...): ...
def truncated_tetrahedron_graph(create_using: Optional[Any] = ...): ...
def tutte_graph(create_using: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.isomorphism.isomorphvf2 (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

class GraphMatcher:
    G1: Any = ...
    G2: Any = ...
    G1_nodes: Any = ...
    G2_nodes: Any = ...
    old_recursion_limit: Any = ...
    test: str = ...
    def __init__(self, G1, G2) -> None: ...
    def reset_recursion_limit(self): ...
    def candidate_pairs_iter(self): ...
    core_1: Any = ...
    core_2: Any = ...
    inout_1: Any = ...
    inout_2: Any = ...
    state: Any = ...
    mapping: Any = ...
    def initialize(self): ...
    def is_isomorphic(self): ...
    def isomorphisms_iter(self): ...
    def match(self): ...
    def semantic_feasibility(self, G1_node, G2_node): ...
    def subgraph_is_isomorphic(self): ...
    def subgraph_isomorphisms_iter(self): ...
    def syntactic_feasibility(self, G1_node, G2_node): ...

class DiGraphMatcher(GraphMatcher):
    def __init__(self, G1, G2) -> None: ...
    def candidate_pairs_iter(self): ...
    core_1: Any = ...
    core_2: Any = ...
    in_1: Any = ...
    in_2: Any = ...
    out_1: Any = ...
    out_2: Any = ...
    state: Any = ...
    mapping: Any = ...
    def initialize(self): ...
    def syntactic_feasibility(self, G1_node, G2_node): ...

class GMState:
    GM: Any = ...
    G1_node: Any = ...
    G2_node: Any = ...
    depth: Any = ...
    def __init__(self, GM, G1_node: Optional[Any] = ..., G2_node: Optional[Any] = ...) -> None: ...
    def restore(self): ...

class DiGMState:
    GM: Any = ...
    G1_node: Any = ...
    G2_node: Any = ...
    depth: Any = ...
    def __init__(self, GM, G1_node: Optional[Any] = ..., G2_node: Optional[Any] = ...) -> None: ...
    def restore(self): ...



# Stubs for networkx.algorithms.flow.edmondskarp (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from networkx.algorithms.flow.utils import *

def edmonds_karp(
    G,
    s,
    t,
    capacity: str = ...,
    residual: Optional[Any] = ...,
    value_only: bool = ...,
    cutoff: Optional[Any] = ...,
): ...



# Stubs for networkx.algorithms.components.weakly_connected (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def weakly_connected_components(G): ...
def number_weakly_connected_components(G): ...
def weakly_connected_component_subgraphs(G, copy: bool = ...): ...
def is_weakly_connected(G): ...



"""Implementation of CommandLineTool."""

import copy
import hashlib
import json
import locale
import logging
import os
import re
import shutil
import threading
import urllib
import urllib.parse
from enum import Enum
from functools import cmp_to_key, partial
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Pattern,
    Set,
    TextIO,
    Type,
    Union,
    cast,
)

import shellescape
from schema_salad.avro.schema import Schema
from schema_salad.exceptions import ValidationException
from schema_salad.ref_resolver import file_uri, uri_file_path
from schema_salad.sourceline import SourceLine
from schema_salad.utils import json_dumps
from schema_salad.validate import validate_ex

from ruamel.yaml.comments import CommentedMap, CommentedSeq

from .builder import (
    INPUT_OBJ_VOCAB,
    Builder,
    content_limit_respected_read_bytes,
    substitute,
)
from .context import LoadingContext, RuntimeContext, getdefault
from .docker import DockerCommandLineJob, PodmanCommandLineJob
from .errors import UnsupportedRequirement, WorkflowException
from .flatten import flatten
from .job import CommandLineJob, JobBase
from .loghandler import _logger
from .mpi import MPIRequirementName
from .mutation import MutationManager
from .pathmapper import PathMapper
from .process import (
    Process,
    _logger_validation_warnings,
    compute_checksums,
    shortname,
    uniquename,
)
from .singularity import SingularityCommandLineJob
from .stdfsaccess import StdFsAccess
from .udocker import UDockerCommandLineJob
from .update import ORDERED_VERSIONS, ORIGINAL_CWLVERSION
from .utils import (
    CWLObjectType,
    CWLOutputType,
    DirectoryType,
    JobsGeneratorType,
    OutputCallbackType,
    adjustDirObjs,
    adjustFileObjs,
    aslist,
    get_listing,
    normalizeFilesDirs,
    random_outdir,
    shared_file_lock,
    trim_listing,
    upgrade_lock,
    visit_class,
)

if TYPE_CHECKING:
    from .provenance_profile import ProvenanceProfile  # pylint: disable=unused-import


class PathCheckingMode(Enum):
    """
    What characters are allowed in path names.

    We have the strict (default) mode and the relaxed mode.
    """

    STRICT = re.compile(r"^[\w.+\,\-:@\]^\u2600-\u26FF\U0001f600-\U0001f64f]+$")
    r"""
    Accepts names that contain one or more of the following:

    .. list-table::

       * - ``\w``
         - unicode word characters

           this includes most characters that can be part of a word in any
           language, as well as numbers and the underscore
       * - ``.``
         - a literal period
       * - ``+``
         - a literal plus sign
       * - ``,``
         - a literal comma
       * - ``-``
         - a literal minus sign
       * - ``:``
         - a literal colon
       * - ``@``
         - a literal at-symbol
       * - ``]``
         - a literal end-square-bracket
       * - ``^``
         - a literal caret symbol
       * - ``\u2600-\u26FF``
         - matches a single character in the range between ☀ (index 9728) and ⛿ (index 9983)
       * - ``\U0001f600-\U0001f64f``
         - matches a single character in the range between 😀 (index 128512) and 🙏 (index 128591)

    Note: the following characters are intentionally not included:

    1. reserved words in POSIX: ``!``, :code:`{`, ``}``

    2. POSIX metacharacters listed in the CWL standard as okay to reject: ``|``,
       ``&``, ``;``, ``<``, ``>``, ``(``, ``)``, ``$``, `````, ``"``, ``'``,
       :kbd:`<space>`, :kbd:`<tab>`, :kbd:`<newline>`.

       (In accordance with https://www.commonwl.org/v1.0/CommandLineTool.html#File under "path" )

    3. POSIX path separator: ``\``

       (also listed at https://www.commonwl.org/v1.0/CommandLineTool.html#File under "path")

    4. Additional POSIX metacharacters: ``*``, ``?``, ``[``, ``#``, ``˜``,
       ``=``, ``%``.

    TODO: switch to https://pypi.org/project/regex/ and use
    ``\p{Extended_Pictographic}`` instead of the manual emoji ranges
    """

    RELAXED = re.compile(r".*")
    """Accept anything."""


class ExpressionJob:
    """Job for :py:class:`ExpressionTool`."""

    def __init__(
        self,
        builder: Builder,
        script: str,
        output_callback: Optional[OutputCallbackType],
        requirements: List[CWLObjectType],
        hints: List[CWLObjectType],
        outdir: Optional[str] = None,
        tmpdir: Optional[str] = None,
    ) -> None:
        """Initialize this ExpressionJob."""
        self.builder = builder
        self.requirements = requirements
        self.hints = hints
        self.output_callback = output_callback
        self.outdir = outdir
        self.tmpdir = tmpdir
        self.script = script
        self.prov_obj: Optional["ProvenanceProfile"] = None

    def run(
        self,
        runtimeContext: RuntimeContext,
        tmpdir_lock: Optional[threading.Lock] = None,
    ) -> None:
        try:
            normalizeFilesDirs(self.builder.job)
            ev = self.builder.do_eval(self.script)
            normalizeFilesDirs(
                cast(
                    Optional[
                        Union[
                            MutableSequence[MutableMapping[str, Any]],
                            MutableMapping[str, Any],
                            DirectoryType,
                        ]
                    ],
                    ev,
                )
            )
            if self.output_callback:
                self.output_callback(cast(Optional[CWLObjectType], ev), "success")
        except WorkflowException as err:
            _logger.warning(
                "Failed to evaluate expression:\n%s",
                str(err),
                exc_info=runtimeContext.debug,
            )
            if self.output_callback:
                self.output_callback({}, "permanentFail")


class ExpressionTool(Process):
    def job(
        self,
        job_order: CWLObjectType,
        output_callbacks: Optional[OutputCallbackType],
        runtimeContext: RuntimeContext,
    ) -> Generator[ExpressionJob, None, None]:
        builder = self._init_job(job_order, runtimeContext)

        job = ExpressionJob(
            builder,
            self.tool["expression"],
            output_callbacks,
            self.requirements,
            self.hints,
        )
        job.prov_obj = runtimeContext.prov_obj
        yield job


class AbstractOperation(Process):
    def job(
        self,
        job_order: CWLObjectType,
        output_callbacks: Optional[OutputCallbackType],
        runtimeContext: RuntimeContext,
    ) -> JobsGeneratorType:
        raise WorkflowException("Abstract operation cannot be executed.")


def remove_path(f):  # type: (CWLObjectType) -> None
    if "path" in f:
        del f["path"]


def revmap_file(builder: Builder, outdir: str, f: CWLObjectType) -> Optional[CWLObjectType]:
    """
    Remap a file from internal path to external path.

    For Docker, this maps from the path inside tho container to the path
    outside the container. Recognizes files in the pathmapper or remaps
    internal output directories to the external directory.
    """
    # builder.outdir is the inner (container/compute node) output directory
    # outdir is the outer (host/storage system) output directory

    if outdir.startswith("/"):
        # local file path, turn it into a file:// URI
        outdir = file_uri(outdir)

    # note: outer outdir should already be a URI and should not be URI
    # quoted any further.

    if "location" in f and "path" not in f:
        location = cast(str, f["location"])
        if location.startswith("file://"):
            f["path"] = uri_file_path(location)
        else:
            f["location"] = builder.fs_access.join(outdir, cast(str, f["location"]))
            return f

    if "dirname" in f:
        del f["dirname"]

    if "path" in f:
        path = builder.fs_access.join(builder.outdir, cast(str, f["path"]))
        uripath = file_uri(path)
        del f["path"]

        if "basename" not in f:
            f["basename"] = os.path.basename(path)

        if not builder.pathmapper:
            raise ValueError(
                "Do not call revmap_file using a builder that doesn't have a pathmapper."
            )
        revmap_f = builder.pathmapper.reversemap(path)

        if revmap_f and not builder.pathmapper.mapper(revmap_f[0]).type.startswith("Writable"):
            f["location"] = revmap_f[1]
        elif (
            uripath == outdir
            or uripath.startswith(outdir + os.sep)
            or uripath.startswith(outdir + "/")
        ):
            f["location"] = uripath
        elif (
            path == builder.outdir
            or path.startswith(builder.outdir + os.sep)
            or path.startswith(builder.outdir + "/")
        ):
            joined_path = builder.fs_access.join(
                outdir, urllib.parse.quote(path[len(builder.outdir) + 1 :])
            )
            f["location"] = joined_path
        else:
            raise WorkflowException(
                "Output file path %s must be within designated output directory (%s) or an input "
                "file pass through." % (path, builder.outdir)
            )
        return f

    raise WorkflowException(
        "Output File object is missing both 'location' and 'path' fields: %s" % f
    )


class CallbackJob:
    """Callback Job class, used by :py:func:`CommandLineTool.job`."""

    def __init__(
        self,
        job: "CommandLineTool",
        output_callback: Optional[OutputCallbackType],
        cachebuilder: Builder,
        jobcache: str,
    ) -> None:
        """Initialize this CallbackJob."""
        self.job = job
        self.output_callback = output_callback
        self.cachebuilder = cachebuilder
        self.outdir = jobcache
        self.prov_obj = None  # type: Optional[ProvenanceProfile]

    def run(
        self,
        runtimeContext: RuntimeContext,
        tmpdir_lock: Optional[threading.Lock] = None,
    ) -> None:
        if self.output_callback:
            self.output_callback(
                self.job.collect_output_ports(
                    self.job.tool["outputs"],
                    self.cachebuilder,
                    self.outdir,
                    getdefault(runtimeContext.compute_checksum, True),
                ),
                "success",
            )


def check_adjust(accept_re: Pattern[str], builder: Builder, file_o: CWLObjectType) -> CWLObjectType:
    """
    Map files to assigned path inside a container.

    We need to also explicitly walk over input, as implicit reassignment
    doesn't reach everything in builder.bindings
    """
    if not builder.pathmapper:
        raise ValueError("Do not call check_adjust using a builder that doesn't have a pathmapper.")
    file_o["path"] = path = builder.pathmapper.mapper(cast(str, file_o["location"]))[1]
    basename = cast(str, file_o.get("basename"))
    dn, bn = os.path.split(path)
    if file_o.get("dirname") != dn:
        file_o["dirname"] = str(dn)
    if basename != bn:
        file_o["basename"] = basename = str(bn)
    if file_o["class"] == "File":
        nr, ne = os.path.splitext(basename)
        if file_o.get("nameroot") != nr:
            file_o["nameroot"] = str(nr)
        if file_o.get("nameext") != ne:
            file_o["nameext"] = str(ne)
    if not accept_re.match(basename):
        raise WorkflowException(
            f"Invalid filename: {file_o['basename']!r} contains illegal characters"
        )
    return file_o


def check_valid_locations(fs_access: StdFsAccess, ob: CWLObjectType) -> None:
    location = cast(str, ob["location"])
    if location.startswith("_:"):
        pass
    if ob["class"] == "File" and not fs_access.isfile(location):
        raise ValidationException("Does not exist or is not a File: '%s'" % location)
    if ob["class"] == "Directory" and not fs_access.isdir(location):
        raise ValidationException("Does not exist or is not a Directory: '%s'" % location)


OutputPortsType = Dict[str, Optional[CWLOutputType]]


class ParameterOutputWorkflowException(WorkflowException):
    def __init__(self, msg: str, port: CWLObjectType, **kwargs: Any) -> None:
        """Exception for when there was an error collecting output for a parameter."""
        super().__init__(
            "Error collecting output for parameter '%s': %s"
            % (shortname(cast(str, port["id"])), msg),
            kwargs,
        )


class CommandLineTool(Process):
    def __init__(self, toolpath_object: CommentedMap, loadingContext: LoadingContext) -> None:
        """Initialize this CommandLineTool."""
        super().__init__(toolpath_object, loadingContext)
        self.prov_obj = loadingContext.prov_obj
        self.path_check_mode = (
            PathCheckingMode.RELAXED
            if loadingContext.relax_path_checks
            else PathCheckingMode.STRICT
        )  # type: PathCheckingMode

    def make_job_runner(self, runtimeContext: RuntimeContext) -> Type[JobBase]:
        dockerReq, dockerRequired = self.get_requirement("DockerRequirement")
        mpiReq, mpiRequired = self.get_requirement(MPIRequirementName)

        if not dockerReq and runtimeContext.use_container:
            if runtimeContext.find_default_container is not None:
                default_container = runtimeContext.find_default_container(self)
                if default_container is not None:
                    dockerReq = {
                        "class": "DockerRequirement",
                        "dockerPull": default_container,
                    }
                    if mpiRequired:
                        self.hints.insert(0, dockerReq)
                        dockerRequired = False
                    else:
                        self.requirements.insert(0, dockerReq)
                        dockerRequired = True

        if dockerReq is not None and runtimeContext.use_container:
            if mpiReq is not None:
                _logger.warning("MPIRequirement with containers is a beta feature")
            if runtimeContext.singularity:
                return SingularityCommandLineJob
            elif runtimeContext.user_space_docker_cmd:
                return UDockerCommandLineJob
            if mpiReq is not None:
                if mpiRequired:
                    if dockerRequired:
                        raise UnsupportedRequirement(
                            "No support for Docker and MPIRequirement both being required"
                        )
                    else:
                        _logger.warning(
                            "MPI has been required while Docker is hinted, discarding Docker hint(s)"
                        )
                        self.hints = [h for h in self.hints if h["class"] != "DockerRequirement"]
                        return CommandLineJob
                else:
                    if dockerRequired:
                        _logger.warning(
                            "Docker has been required while MPI is hinted, discarding MPI hint(s)"
                        )
                        self.hints = [h for h in self.hints if h["class"] != MPIRequirementName]
                    else:
                        raise UnsupportedRequirement(
                            "Both Docker and MPI have been hinted - don't know what to do"
                        )
            if runtimeContext.podman:
                return PodmanCommandLineJob
            return DockerCommandLineJob
        if dockerRequired:
            raise UnsupportedRequirement(
                "--no-container, but this CommandLineTool has "
                "DockerRequirement under 'requirements'."
            )
        return CommandLineJob

    def make_path_mapper(
        self,
        reffiles: List[CWLObjectType],
        stagedir: str,
        runtimeContext: RuntimeContext,
        separateDirs: bool,
    ) -> PathMapper:
        return PathMapper(reffiles, runtimeContext.basedir, stagedir, separateDirs)

    def updatePathmap(self, outdir: str, pathmap: PathMapper, fn: CWLObjectType) -> None:
        """Update a PathMapper with a CWL File or Directory object."""
        if not isinstance(fn, MutableMapping):
            raise WorkflowException("Expected File or Directory object, was %s" % type(fn))
        basename = cast(str, fn["basename"])
        if "location" in fn:
            location = cast(str, fn["location"])
            if location in pathmap:
                pathmap.update(
                    location,
                    pathmap.mapper(location).resolved,
                    os.path.join(outdir, basename),
                    ("Writable" if fn.get("writable") else "") + cast(str, fn["class"]),
                    False,
                )
        for sf in cast(List[CWLObjectType], fn.get("secondaryFiles", [])):
            self.updatePathmap(outdir, pathmap, sf)
        for ls in cast(List[CWLObjectType], fn.get("listing", [])):
            self.updatePathmap(os.path.join(outdir, cast(str, fn["basename"])), pathmap, ls)

    def _initialworkdir(self, j: JobBase, builder: Builder) -> None:
        initialWorkdir, _ = self.get_requirement("InitialWorkDirRequirement")
        if initialWorkdir is None:
            return
        debug = _logger.isEnabledFor(logging.DEBUG)
        cwl_version = cast(Optional[str], self.metadata.get(ORIGINAL_CWLVERSION, None))
        classic_dirent: bool = cwl_version is not None and (
            ORDERED_VERSIONS.index(cwl_version) < ORDERED_VERSIONS.index("v1.2.0-dev2")
        )
        classic_listing = cwl_version and ORDERED_VERSIONS.index(
            cwl_version
        ) < ORDERED_VERSIONS.index("v1.1.0-dev1")

        ls = []  # type: List[CWLObjectType]
        if isinstance(initialWorkdir["listing"], str):
            # "listing" is just a string (must be an expression) so
            # just evaluate it and use the result as if it was in
            # listing
            ls_evaluated = builder.do_eval(initialWorkdir["listing"])
            fail: Any = False
            fail_suffix: str = ""
            if not isinstance(ls_evaluated, MutableSequence):
                fail = ls_evaluated
            else:
                ls_evaluated2 = cast(MutableSequence[Union[None, CWLOutputType]], ls_evaluated)
                for entry in ls_evaluated2:
                    if entry == None:  # noqa
                        if classic_dirent:
                            fail = entry
                            fail_suffix = (
                                " Dirent.entry cannot return 'null' before CWL "
                                "v1.2. Please consider using 'cwl-upgrader' to "
                                "upgrade your document to CWL version v1.2."
                            )
                    elif isinstance(entry, MutableSequence):
                        if classic_listing:
                            raise SourceLine(
                                initialWorkdir, "listing", WorkflowException, debug
                            ).makeError(
                                "InitialWorkDirRequirement.listing expressions "
                                "cannot return arrays of Files or Directories "
                                "before CWL v1.1. Please "
                                "considering using 'cwl-upgrader' to upgrade "
                                "your document to CWL v1.1' or later."
                            )
                        else:
                            for entry2 in entry:
                                if not (
                                    isinstance(entry2, MutableMapping)
                                    and (
                                        "class" in entry2
                                        and entry2["class"] == "File"
                                        or "Directory"
                                    )
                                ):
                                    fail = (
                                        "an array with an item ('{entry2}') that is "
                                        "not a File nor a Directory object."
                                    )
                    elif not (
                        isinstance(entry, MutableMapping)
                        and (
                            "class" in entry
                            and (entry["class"] == "File" or "Directory")
                            or "entry" in entry
                        )
                    ):
                        fail = entry
            if fail is not False:
                message = (
                    "Expression in a 'InitialWorkdirRequirement.listing' field "
                    "must return a list containing zero or more of: File or "
                    "Directory objects; Dirent objects"
                )
                if classic_dirent:
                    message += ". "
                else:
                    message += "; null; or arrays of File or Directory objects. "
                message += f"Got {fail!r} among the results from "
                message += f"{initialWorkdir['listing'].strip()!r}." + fail_suffix
                raise SourceLine(initialWorkdir, "listing", WorkflowException, debug).makeError(
                    message
                )
            ls = cast(List[CWLObjectType], ls_evaluated)
        else:
            # "listing" is an array of either expressions or Dirent so
            # evaluate each item
            for t in cast(
                MutableSequence[Union[str, CWLObjectType]],
                initialWorkdir["listing"],
            ):
                if isinstance(t, Mapping) and "entry" in t:
                    # Dirent
                    entry_field = cast(str, t["entry"])
                    # the schema guarantees that 'entry' is a string, so the cast is safe
                    entry = builder.do_eval(entry_field, strip_whitespace=False)
                    if entry is None:
                        continue

                    if isinstance(entry, MutableSequence):
                        if classic_listing:
                            raise SourceLine(t, "entry", WorkflowException, debug).makeError(
                                "'entry' expressions are not allowed to evaluate "
                                "to an array of Files or Directories until CWL "
                                "v1.2. Consider using 'cwl-upgrader' to upgrade "
                                "your document to CWL version 1.2."
                            )
                        # Nested list.  If it is a list of File or
                        # Directory objects, add it to the
                        # file list, otherwise JSON serialize it if CWL v1.2.

                        filelist = True
                        for e in entry:
                            if not isinstance(e, MutableMapping) or e.get("class") not in (
                                "File",
                                "Directory",
                            ):
                                filelist = False
                                break

                        if filelist:
                            if "entryname" in t:
                                raise SourceLine(
                                    t, "entryname", WorkflowException, debug
                                ).makeError(
                                    "'entryname' is invalid when 'entry' returns list of File or Directory"
                                )
                            for e in entry:
                                ec = cast(CWLObjectType, e)
                                ec["writable"] = t.get("writable", False)
                            ls.extend(cast(List[CWLObjectType], entry))
                            continue

                    et = {}  # type: CWLObjectType
                    if isinstance(entry, Mapping) and entry.get("class") in (
                        "File",
                        "Directory",
                    ):
                        et["entry"] = cast(CWLOutputType, entry)
                    else:
                        if isinstance(entry, str):
                            et["entry"] = entry
                        else:
                            if classic_dirent:
                                raise SourceLine(t, "entry", WorkflowException, debug).makeError(
                                    "'entry' expression resulted in "
                                    "something other than number, object or "
                                    "array besides a single File or Dirent object. "
                                    "In CWL v1.2+ this would be serialized to a JSON object. "
                                    "However this is a {cwl_version} document. "
                                    "If that is the desired result then please "
                                    "consider using 'cwl-upgrader' to upgrade "
                                    "your document to CWL version 1.2. "
                                    f"Result of {entry_field!r} was {entry!r}."
                                )
                            et["entry"] = json_dumps(entry, sort_keys=True)

                    if "entryname" in t:
                        entryname_field = cast(str, t["entryname"])
                        if "${" in entryname_field or "$(" in entryname_field:
                            en = builder.do_eval(cast(str, t["entryname"]))
                            if not isinstance(en, str):
                                raise SourceLine(
                                    t, "entryname", WorkflowException, debug
                                ).makeError(
                                    "'entryname' expression must result a string. "
                                    f"Got {en!r} from {entryname_field!r}"
                                )
                            et["entryname"] = en
                        else:
                            et["entryname"] = entryname_field
                    else:
                        et["entryname"] = None
                    et["writable"] = t.get("writable", False)
                    ls.append(et)
                else:
                    # Expression, must return a Dirent, File, Directory
                    # or array of such.
                    initwd_item = builder.do_eval(t)
                    if not initwd_item:
                        continue
                    if isinstance(initwd_item, MutableSequence):
                        ls.extend(cast(List[CWLObjectType], initwd_item))
                    else:
                        ls.append(cast(CWLObjectType, initwd_item))

        for i, t2 in enumerate(ls):
            if not isinstance(t2, Mapping):
                raise SourceLine(initialWorkdir, "listing", WorkflowException, debug).makeError(
                    "Entry at index %s of listing is not a record, was %s" % (i, type(t2))
                )

            if "entry" not in t2:
                continue

            # Dirent
            if isinstance(t2["entry"], str):
                if not t2["entryname"]:
                    raise SourceLine(initialWorkdir, "listing", WorkflowException, debug).makeError(
                        "Entry at index %s of listing missing entryname" % (i)
                    )
                ls[i] = {
                    "class": "File",
                    "basename": t2["entryname"],
                    "contents": t2["entry"],
                    "writable": t2.get("writable"),
                }
                continue

            if not isinstance(t2["entry"], Mapping):
                raise SourceLine(initialWorkdir, "listing", WorkflowException, debug).makeError(
                    "Entry at index %s of listing is not a record, was %s" % (i, type(t2["entry"]))
                )

            if t2["entry"].get("class") not in ("File", "Directory"):
                raise SourceLine(initialWorkdir, "listing", WorkflowException, debug).makeError(
                    "Entry at index %s of listing is not a File or Directory object, was %s"
                    % (i, t2)
                )

            if t2.get("entryname") or t2.get("writable"):
                t2 = copy.deepcopy(t2)
                t2entry = cast(CWLObjectType, t2["entry"])
                if t2.get("entryname"):
                    t2entry["basename"] = t2["entryname"]
                t2entry["writable"] = t2.get("writable")

            ls[i] = cast(CWLObjectType, t2["entry"])

        for i, t3 in enumerate(ls):
            if t3.get("class") not in ("File", "Directory"):
                # Check that every item is a File or Directory object now
                raise SourceLine(initialWorkdir, "listing", WorkflowException, debug).makeError(
                    f"Entry at index {i} of listing is not a Dirent, File or "
                    f"Directory object, was {t2}."
                )
            if "basename" not in t3:
                continue
            basename = os.path.normpath(cast(str, t3["basename"]))
            t3["basename"] = basename
            if basename.startswith("../"):
                raise SourceLine(initialWorkdir, "listing", WorkflowException, debug).makeError(
                    f"Name {basename!r} at index {i} of listing is invalid, "
                    "cannot start with '../'"
                )
            if basename.startswith("/"):
                # only if DockerRequirement in requirements
                if cwl_version and ORDERED_VERSIONS.index(cwl_version) < ORDERED_VERSIONS.index(
                    "v1.2.0-dev4"
                ):
                    raise SourceLine(initialWorkdir, "listing", WorkflowException, debug).makeError(
                        f"Name {basename!r} at index {i} of listing is invalid, "
                        "paths starting with '/' are only permitted in CWL 1.2 "
                        "and later. Consider changing the absolute path to a relative "
                        "path, or upgrade the CWL description to CWL v1.2 using "
                        "https://pypi.org/project/cwl-upgrader/"
                    )

                req, is_req = self.get_requirement("DockerRequirement")
                if is_req is not True:
                    raise SourceLine(initialWorkdir, "listing", WorkflowException, debug).makeError(
                        f"Name {basename!r} at index {i} of listing is invalid, "
                        "name can only start with '/' when DockerRequirement "
                        "is in 'requirements'."
                    )

        with SourceLine(initialWorkdir, "listing", WorkflowException, debug):
            j.generatefiles["listing"] = ls
            for entry in ls:
                if "basename" in entry:
                    basename = cast(str, entry["basename"])
                    entry["dirname"] = os.path.join(builder.outdir, os.path.dirname(basename))
                    entry["basename"] = os.path.basename(basename)
                normalizeFilesDirs(entry)
                self.updatePathmap(
                    cast(Optional[str], entry.get("dirname")) or builder.outdir,
                    cast(PathMapper, builder.pathmapper),
                    entry,
                )
                if "listing" in entry:

                    def remove_dirname(d: CWLObjectType) -> None:
                        if "dirname" in d:
                            del d["dirname"]

                    visit_class(
                        entry["listing"],
                        ("File", "Directory"),
                        remove_dirname,
                    )

            visit_class(
                [builder.files, builder.bindings],
                ("File", "Directory"),
                partial(check_adjust, self.path_check_mode.value, builder),
            )

    def job(
        self,
        job_order: CWLObjectType,
        output_callbacks: Optional[OutputCallbackType],
        runtimeContext: RuntimeContext,
    ) -> Generator[Union[JobBase, CallbackJob], None, None]:
        workReuse, _ = self.get_requirement("WorkReuse")
        enableReuse = workReuse.get("enableReuse", True) if workReuse else True

        jobname = uniquename(runtimeContext.name or shortname(self.tool.get("id", "job")))
        if runtimeContext.cachedir and enableReuse:
            cachecontext = runtimeContext.copy()
            cachecontext.outdir = "/out"
            cachecontext.tmpdir = "/tmp"  # nosec
            cachecontext.stagedir = "/stage"
            cachebuilder = self._init_job(job_order, cachecontext)
            cachebuilder.pathmapper = PathMapper(
                cachebuilder.files,
                runtimeContext.basedir,
                cachebuilder.stagedir,
                separateDirs=False,
            )
            _check_adjust = partial(check_adjust, self.path_check_mode.value, cachebuilder)
            _checksum = partial(
                compute_checksums,
                runtimeContext.make_fs_access(runtimeContext.basedir),
            )
            visit_class(
                [cachebuilder.files, cachebuilder.bindings],
                ("File", "Directory"),
                _check_adjust,
            )
            visit_class([cachebuilder.files, cachebuilder.bindings], ("File"), _checksum)

            cmdline = flatten(list(map(cachebuilder.generate_arg, cachebuilder.bindings)))
            docker_req, _ = self.get_requirement("DockerRequirement")
            if docker_req is not None and runtimeContext.use_container:
                dockerimg = docker_req.get("dockerImageId") or docker_req.get("dockerPull")
            elif runtimeContext.default_container is not None and runtimeContext.use_container:
                dockerimg = runtimeContext.default_container
            else:
                dockerimg = None

            if dockerimg is not None:
                cmdline = ["docker", "run", dockerimg] + cmdline
                # not really run using docker, just for hashing purposes

            keydict = {
                "cmdline": cmdline
            }  # type: Dict[str, Union[MutableSequence[Union[str, int]], CWLObjectType]]

            for shortcut in ["stdin", "stdout", "stderr"]:
                if shortcut in self.tool:
                    keydict[shortcut] = self.tool[shortcut]

            def calc_checksum(location: str) -> Optional[str]:
                for e in cachebuilder.files:
                    if (
                        "location" in e
                        and e["location"] == location
                        and "checksum" in e
                        and e["checksum"] != "sha1$hash"
                    ):
                        return cast(Optional[str], e["checksum"])
                return None

            def remove_prefix(s: str, prefix: str) -> str:
                # replace with str.removeprefix when Python 3.9+
                return s[len(prefix) :] if s.startswith(prefix) else s

            for location, fobj in cachebuilder.pathmapper.items():
                if fobj.type == "File":
                    checksum = calc_checksum(location)
                    fobj_stat = os.stat(fobj.resolved)
                    path = remove_prefix(fobj.resolved, runtimeContext.basedir + "/")
                    if checksum is not None:
                        keydict[path] = [fobj_stat.st_size, checksum]
                    else:
                        keydict[path] = [
                            fobj_stat.st_size,
                            int(fobj_stat.st_mtime * 1000),
                        ]

            interesting = {
                "DockerRequirement",
                "EnvVarRequirement",
                "InitialWorkDirRequirement",
                "ShellCommandRequirement",
                "NetworkAccess",
            }
            for rh in (self.original_requirements, self.original_hints):
                for r in reversed(rh):
                    cls = cast(str, r["class"])
                    if cls in interesting and cls not in keydict:
                        keydict[cls] = r

            keydictstr = json_dumps(keydict, separators=(",", ":"), sort_keys=True)
            cachekey = hashlib.md5(keydictstr.encode("utf-8")).hexdigest()  # nosec

            _logger.debug("[job %s] keydictstr is %s -> %s", jobname, keydictstr, cachekey)

            jobcache = os.path.join(runtimeContext.cachedir, cachekey)

            # Create a lockfile to manage cache status.
            jobcachepending = f"{jobcache}.status"
            jobcachelock = None
            jobstatus = None

            # Opens the file for read/write, or creates an empty file.
            jobcachelock = open(jobcachepending, "a+")

            # get the shared lock to ensure no other process is trying
            # to write to this cache
            shared_file_lock(jobcachelock)
            jobcachelock.seek(0)
            jobstatus = jobcachelock.read()

            if os.path.isdir(jobcache) and jobstatus == "success":
                if docker_req and runtimeContext.use_container:
                    cachebuilder.outdir = runtimeContext.docker_outdir or random_outdir()
                else:
                    cachebuilder.outdir = jobcache

                _logger.info("[job %s] Using cached output in %s", jobname, jobcache)
                yield CallbackJob(self, output_callbacks, cachebuilder, jobcache)
                # we're done with the cache so release lock
                jobcachelock.close()
                return
            else:
                _logger.info("[job %s] Output of job will be cached in %s", jobname, jobcache)

                # turn shared lock into an exclusive lock since we'll
                # be writing the cache directory
                upgrade_lock(jobcachelock)

                shutil.rmtree(jobcache, True)
                os.makedirs(jobcache)
                runtimeContext = runtimeContext.copy()
                runtimeContext.outdir = jobcache

                def update_status_output_callback(
                    output_callbacks: OutputCallbackType,
                    jobcachelock: TextIO,
                    outputs: Optional[CWLObjectType],
                    processStatus: str,
                ) -> None:
                    # save status to the lockfile then release the lock
                    jobcachelock.seek(0)
                    jobcachelock.truncate()
                    jobcachelock.write(processStatus)
                    jobcachelock.close()
                    output_callbacks(outputs, processStatus)

                output_callbacks = partial(
                    update_status_output_callback, output_callbacks, jobcachelock
                )

        builder = self._init_job(job_order, runtimeContext)

        reffiles = copy.deepcopy(builder.files)

        j = self.make_job_runner(runtimeContext)(
            builder,
            builder.job,
            self.make_path_mapper,
            self.requirements,
            self.hints,
            jobname,
        )
        j.prov_obj = self.prov_obj

        j.successCodes = self.tool.get("successCodes", [])
        j.temporaryFailCodes = self.tool.get("temporaryFailCodes", [])
        j.permanentFailCodes = self.tool.get("permanentFailCodes", [])

        debug = _logger.isEnabledFor(logging.DEBUG)

        if debug:
            _logger.debug(
                "[job %s] initializing from %s%s",
                j.name,
                self.tool.get("id", ""),
                " as part of %s" % runtimeContext.part_of if runtimeContext.part_of else "",
            )
            _logger.debug("[job %s] %s", j.name, json_dumps(builder.job, indent=4))

        builder.pathmapper = self.make_path_mapper(reffiles, builder.stagedir, runtimeContext, True)
        builder.requirements = j.requirements

        _check_adjust = partial(check_adjust, self.path_check_mode.value, builder)

        visit_class([builder.files, builder.bindings], ("File", "Directory"), _check_adjust)

        self._initialworkdir(j, builder)

        if debug:
            _logger.debug(
                "[job %s] path mappings is %s",
                j.name,
                json_dumps(
                    {p: builder.pathmapper.mapper(p) for p in builder.pathmapper.files()},
                    indent=4,
                ),
            )

        if self.tool.get("stdin"):
            with SourceLine(self.tool, "stdin", ValidationException, debug):
                stdin_eval = builder.do_eval(self.tool["stdin"])
                if not (isinstance(stdin_eval, str) or stdin_eval is None):
                    raise ValidationException(
                        f"'stdin' expression must return a string or null. Got {stdin_eval!r} "
                        f"for {self.tool['stdin']!r}."
                    )
                j.stdin = stdin_eval
                if j.stdin:
                    reffiles.append({"class": "File", "path": j.stdin})

        if self.tool.get("stderr"):
            with SourceLine(self.tool, "stderr", ValidationException, debug):
                stderr_eval = builder.do_eval(self.tool["stderr"])
                if not isinstance(stderr_eval, str):
                    raise ValidationException(
                        f"'stderr' expression must return a string. Got {stderr_eval!r} "
                        f"for {self.tool['stderr']!r}."
                    )
                j.stderr = stderr_eval
                if j.stderr:
                    if os.path.isabs(j.stderr) or ".." in j.stderr:
                        raise ValidationException(
                            "stderr must be a relative path, got '%s'" % j.stderr
                        )

        if self.tool.get("stdout"):
            with SourceLine(self.tool, "stdout", ValidationException, debug):
                stdout_eval = builder.do_eval(self.tool["stdout"])
                if not isinstance(stdout_eval, str):
                    raise ValidationException(
                        f"'stdout' expression must return a string. Got {stdout_eval!r} "
                        f"for {self.tool['stdout']!r}."
                    )
                j.stdout = stdout_eval
                if j.stdout:
                    if os.path.isabs(j.stdout) or ".." in j.stdout or not j.stdout:
                        raise ValidationException(
                            "stdout must be a relative path, got '%s'" % j.stdout
                        )

        if debug:
            _logger.debug(
                "[job %s] command line bindings is %s",
                j.name,
                json_dumps(builder.bindings, indent=4),
            )
        dockerReq, _ = self.get_requirement("DockerRequirement")
        if dockerReq is not None and runtimeContext.use_container:
            j.outdir = runtimeContext.get_outdir()
            j.tmpdir = runtimeContext.get_tmpdir()
            j.stagedir = runtimeContext.create_tmpdir()
        else:
            j.outdir = builder.outdir
            j.tmpdir = builder.tmpdir
            j.stagedir = builder.stagedir

        inplaceUpdateReq, _ = self.get_requirement("InplaceUpdateRequirement")
        if inplaceUpdateReq is not None:
            j.inplace_update = cast(bool, inplaceUpdateReq["inplaceUpdate"])
        normalizeFilesDirs(j.generatefiles)

        readers = {}  # type: Dict[str, CWLObjectType]
        muts = set()  # type: Set[str]

        if builder.mutation_manager is not None:

            def register_mut(f: CWLObjectType) -> None:
                mm = cast(MutationManager, builder.mutation_manager)
                muts.add(cast(str, f["location"]))
                mm.register_mutation(j.name, f)

            def register_reader(f: CWLObjectType) -> None:
                mm = cast(MutationManager, builder.mutation_manager)
                if cast(str, f["location"]) not in muts:
                    mm.register_reader(j.name, f)
                    readers[cast(str, f["location"])] = copy.deepcopy(f)

            for li in j.generatefiles["listing"]:
                if li.get("writable") and j.inplace_update:
                    adjustFileObjs(li, register_mut)
                    adjustDirObjs(li, register_mut)
                else:
                    adjustFileObjs(li, register_reader)
                    adjustDirObjs(li, register_reader)

            adjustFileObjs(builder.files, register_reader)
            adjustFileObjs(builder.bindings, register_reader)
            adjustDirObjs(builder.files, register_reader)
            adjustDirObjs(builder.bindings, register_reader)

        timelimit, _ = self.get_requirement("ToolTimeLimit")
        if timelimit is not None:
            with SourceLine(timelimit, "timelimit", ValidationException, debug):
                limit_field = cast(Dict[str, Union[str, int]], timelimit)["timelimit"]
                if isinstance(limit_field, str):
                    timelimit_eval = builder.do_eval(limit_field)
                    if timelimit_eval and not isinstance(timelimit_eval, int):
                        raise WorkflowException(
                            "'timelimit' expression must evaluate to a long/int. Got "
                            f"{timelimit_eval!r} for expression {limit_field!r}."
                        )
                else:
                    timelimit_eval = limit_field
                if not isinstance(timelimit_eval, int) or timelimit_eval < 0:
                    raise WorkflowException(
                        f"timelimit must be an integer >= 0, got: {timelimit_eval!r}"
                    )
                j.timelimit = timelimit_eval

        networkaccess, _ = self.get_requirement("NetworkAccess")
        if networkaccess is not None:
            with SourceLine(networkaccess, "networkAccess", ValidationException, debug):
                networkaccess_field = networkaccess["networkAccess"]
                if isinstance(networkaccess_field, str):
                    networkaccess_eval = builder.do_eval(networkaccess_field)
                    if not isinstance(networkaccess_eval, bool):
                        raise WorkflowException(
                            "'networkAccess' expression must evaluate to a bool. "
                            f"Got {networkaccess_eval!r} for expression {networkaccess_field!r}."
                        )
                else:
                    networkaccess_eval = networkaccess_field
                if not isinstance(networkaccess_eval, bool):
                    raise WorkflowException(
                        "networkAccess must be a boolean, got: {networkaccess_eval!r}."
                    )
                j.networkaccess = networkaccess_eval

        # Build a mapping to hold any EnvVarRequirement
        required_env = {}
        evr, _ = self.get_requirement("EnvVarRequirement")
        if evr is not None:
            for eindex, t3 in enumerate(cast(List[Dict[str, str]], evr["envDef"])):
                env_value_field = t3["envValue"]
                if "${" in env_value_field or "$(" in env_value_field:
                    env_value_eval = builder.do_eval(env_value_field)
                    if not isinstance(env_value_eval, str):
                        raise SourceLine(evr["envDef"], eindex, WorkflowException, debug).makeError(
                            "'envValue expression must evaluate to a str. "
                            f"Got {env_value_eval!r} for expression {env_value_field!r}."
                        )
                    env_value = env_value_eval
                else:
                    env_value = env_value_field
                required_env[t3["envName"]] = env_value
        # Construct the env
        j.prepare_environment(runtimeContext, required_env)

        shellcmd, _ = self.get_requirement("ShellCommandRequirement")
        if shellcmd is not None:
            cmd = []  # type: List[str]
            for b in builder.bindings:
                arg = builder.generate_arg(b)
                if b.get("shellQuote", True):
                    arg = [shellescape.quote(a) for a in aslist(arg)]
                cmd.extend(aslist(arg))
            j.command_line = ["/bin/sh", "-c", " ".join(cmd)]
        else:
            j.command_line = flatten(list(map(builder.generate_arg, builder.bindings)))

        j.pathmapper = builder.pathmapper
        j.collect_outputs = partial(
            self.collect_output_ports,
            self.tool["outputs"],
            builder,
            compute_checksum=getdefault(runtimeContext.compute_checksum, True),
            jobname=jobname,
            readers=readers,
        )
        j.output_callback = output_callbacks

        mpi, _ = self.get_requirement(MPIRequirementName)

        if mpi is not None:
            np = cast(  # From the schema for MPIRequirement.processes
                Union[int, str],
                mpi.get("processes", runtimeContext.mpi_config.default_nproc),
            )
            if isinstance(np, str):
                np_eval = builder.do_eval(np)
                if not isinstance(np_eval, int):
                    raise SourceLine(mpi, "processes", WorkflowException, debug).makeError(
                        f"{MPIRequirementName} needs 'processes' expression to "
                        f"evaluate to an int, got {np_eval!r} for expression {np!r}."
                    )
                np = np_eval
            j.mpi_procs = np
        yield j

    def collect_output_ports(
        self,
        ports: Union[CommentedSeq, Set[CWLObjectType]],
        builder: Builder,
        outdir: str,
        rcode: int,
        compute_checksum: bool = True,
        jobname: str = "",
        readers: Optional[MutableMapping[str, CWLObjectType]] = None,
    ) -> OutputPortsType:
        ret = {}  # type: OutputPortsType
        debug = _logger.isEnabledFor(logging.DEBUG)
        cwl_version = self.metadata.get(ORIGINAL_CWLVERSION, None)
        if cwl_version != "v1.0":
            builder.resources["exitCode"] = rcode
        try:
            fs_access = builder.make_fs_access(outdir)
            custom_output = fs_access.join(outdir, "cwl.output.json")
            if fs_access.exists(custom_output):
                with fs_access.open(custom_output, "r") as f:
                    ret = json.load(f)
                if debug:
                    _logger.debug(
                        "Raw output from %s: %s",
                        custom_output,
                        json_dumps(ret, indent=4),
                    )
            else:
                for i, port in enumerate(ports):
                    with SourceLine(
                        ports,
                        i,
                        partial(ParameterOutputWorkflowException, port=port),
                        debug,
                    ):
                        fragment = shortname(port["id"])
                        ret[fragment] = self.collect_output(
                            port,
                            builder,
                            outdir,
                            fs_access,
                            compute_checksum=compute_checksum,
                        )
            if ret:
                revmap = partial(revmap_file, builder, outdir)
                adjustDirObjs(ret, trim_listing)
                visit_class(ret, ("File", "Directory"), revmap)
                visit_class(ret, ("File", "Directory"), remove_path)
                normalizeFilesDirs(ret)
                visit_class(
                    ret,
                    ("File", "Directory"),
                    partial(check_valid_locations, fs_access),
                )

                if compute_checksum:
                    adjustFileObjs(ret, partial(compute_checksums, fs_access))
            expected_schema = cast(Schema, self.names.get_name("outputs_record_schema", None))
            validate_ex(
                expected_schema,
                ret,
                strict=False,
                logger=_logger_validation_warnings,
                vocab=INPUT_OBJ_VOCAB,
            )
            if ret is not None and builder.mutation_manager is not None:
                adjustFileObjs(ret, builder.mutation_manager.set_generation)
            return ret if ret is not None else {}
        except ValidationException as e:
            raise WorkflowException(
                "Error validating output record. " + str(e) + "\n in " + json_dumps(ret, indent=4)
            ) from e
        finally:
            if builder.mutation_manager and readers:
                for r in readers.values():
                    builder.mutation_manager.release_reader(jobname, r)

    def collect_output(
        self,
        schema: CWLObjectType,
        builder: Builder,
        outdir: str,
        fs_access: StdFsAccess,
        compute_checksum: bool = True,
    ) -> Optional[CWLOutputType]:
        r = []  # type: List[CWLOutputType]
        empty_and_optional = False
        debug = _logger.isEnabledFor(logging.DEBUG)
        result: Optional[CWLOutputType] = None
        if "outputBinding" in schema:
            binding = cast(
                MutableMapping[str, Union[bool, str, List[str]]],
                schema["outputBinding"],
            )
            globpatterns = []  # type: List[str]

            revmap = partial(revmap_file, builder, outdir)

            if "glob" in binding:
                with SourceLine(binding, "glob", WorkflowException, debug):
                    for gb in aslist(binding["glob"]):
                        gb = builder.do_eval(gb)
                        if gb:
                            gb_eval_fail = False
                            if not isinstance(gb, str):
                                if isinstance(gb, list):
                                    for entry in gb:
                                        if not isinstance(entry, str):
                                            gb_eval_fail = True
                                else:
                                    gb_eval_fail = True
                            if gb_eval_fail:
                                raise WorkflowException(
                                    "Resolved glob patterns must be strings "
                                    f"or list of strings, not "
                                    f"{gb!r} from {binding['glob']!r}"
                                )
                            globpatterns.extend(aslist(gb))

                    for gb in globpatterns:
                        if gb.startswith(builder.outdir):
                            gb = gb[len(builder.outdir) + 1 :]
                        elif gb == ".":
                            gb = outdir
                        elif gb.startswith("/"):
                            raise WorkflowException("glob patterns must not start with '/'")
                        try:
                            prefix = fs_access.glob(outdir)
                            sorted_glob_result = sorted(
                                fs_access.glob(fs_access.join(outdir, gb)),
                                key=cmp_to_key(locale.strcoll),
                            )
                            r.extend(
                                [
                                    {
                                        "location": g,
                                        "path": fs_access.join(
                                            builder.outdir,
                                            urllib.parse.unquote(g[len(prefix[0]) + 1 :]),
                                        ),
                                        "basename": decoded_basename,
                                        "nameroot": os.path.splitext(decoded_basename)[0],
                                        "nameext": os.path.splitext(decoded_basename)[1],
                                        "class": "File" if fs_access.isfile(g) else "Directory",
                                    }
                                    for g, decoded_basename in zip(
                                        sorted_glob_result,
                                        map(
                                            lambda x: os.path.basename(urllib.parse.unquote(x)),
                                            sorted_glob_result,
                                        ),
                                    )
                                ]
                            )
                        except OSError as e:
                            _logger.warning(str(e))
                        except Exception:
                            _logger.error("Unexpected error from fs_access", exc_info=True)
                            raise

                for files in cast(List[Dict[str, Optional[CWLOutputType]]], r):
                    rfile = files.copy()
                    revmap(rfile)
                    if files["class"] == "Directory":
                        ll = binding.get("loadListing") or builder.loadListing
                        if ll and ll != "no_listing":
                            get_listing(fs_access, files, (ll == "deep_listing"))
                    else:
                        if binding.get("loadContents"):
                            with fs_access.open(cast(str, rfile["location"]), "rb") as f:
                                files["contents"] = content_limit_respected_read_bytes(f).decode(
                                    "utf-8"
                                )
                        if compute_checksum:
                            with fs_access.open(cast(str, rfile["location"]), "rb") as f:
                                checksum = hashlib.sha1()  # nosec
                                contents = f.read(1024 * 1024)
                                while contents != b"":
                                    checksum.update(contents)
                                    contents = f.read(1024 * 1024)
                                files["checksum"] = "sha1$%s" % checksum.hexdigest()
                        files["size"] = fs_access.size(cast(str, rfile["location"]))

            optional = False
            single = False
            if isinstance(schema["type"], MutableSequence):
                if "null" in schema["type"]:
                    optional = True
                if "File" in schema["type"] or "Directory" in schema["type"]:
                    single = True
            elif schema["type"] == "File" or schema["type"] == "Directory":
                single = True

            if "outputEval" in binding:
                with SourceLine(binding, "outputEval", WorkflowException, debug):
                    result = builder.do_eval(cast(CWLOutputType, binding["outputEval"]), context=r)
            else:
                result = cast(CWLOutputType, r)

            if single:
                with SourceLine(binding, "glob", WorkflowException, debug):
                    if not result and not optional:
                        raise WorkflowException(
                            f"Did not find output file with glob pattern: {globpatterns!r}."
                        )
                    elif not result and optional:
                        pass
                    elif isinstance(result, MutableSequence):
                        if len(result) > 1:
                            raise WorkflowException(
                                "Multiple matches for output item that is a single file."
                            )
                        else:
                            result = cast(CWLOutputType, result[0])

            if "secondaryFiles" in schema:
                with SourceLine(schema, "secondaryFiles", WorkflowException, debug):
                    for primary in aslist(result):
                        if isinstance(primary, MutableMapping):
                            primary.setdefault("secondaryFiles", [])
                            pathprefix = primary["path"][0 : primary["path"].rindex(os.sep) + 1]
                            for sf in aslist(schema["secondaryFiles"]):
                                if "required" in sf:
                                    with SourceLine(
                                        schema["secondaryFiles"],
                                        "required",
                                        WorkflowException,
                                        debug,
                                    ):
                                        sf_required_eval = builder.do_eval(
                                            sf["required"], context=primary
                                        )
                                        if not (
                                            isinstance(sf_required_eval, bool)
                                            or sf_required_eval is None
                                        ):
                                            raise WorkflowException(
                                                "Expressions in the field "
                                                "'required' must evaluate to a "
                                                "Boolean (true or false) or None. "
                                                f"Got {sf_required_eval!r} for "
                                                f"{sf['required']!r}."
                                            )
                                        sf_required: bool = sf_required_eval or False
                                else:
                                    sf_required = False

                                if "$(" in sf["pattern"] or "${" in sf["pattern"]:
                                    sfpath = builder.do_eval(sf["pattern"], context=primary)
                                else:
                                    sfpath = substitute(primary["basename"], sf["pattern"])

                                for sfitem in aslist(sfpath):
                                    if not sfitem:
                                        continue
                                    if isinstance(sfitem, str):
                                        sfitem = {"path": pathprefix + sfitem}
                                    if not fs_access.exists(sfitem["path"]) and sf_required:
                                        raise WorkflowException(
                                            "Missing required secondary file '%s'"
                                            % (sfitem["path"])
                                        )
                                    if "path" in sfitem and "location" not in sfitem:
                                        revmap(sfitem)
                                    if fs_access.isfile(sfitem["location"]):
                                        sfitem["class"] = "File"
                                        primary["secondaryFiles"].append(sfitem)
                                    elif fs_access.isdir(sfitem["location"]):
                                        sfitem["class"] = "Directory"
                                        primary["secondaryFiles"].append(sfitem)

            if "format" in schema:
                format_field = cast(str, schema["format"])
                if "$(" in format_field or "${" in format_field:
                    for index, primary in enumerate(aslist(result)):
                        format_eval = builder.do_eval(format_field, context=primary)
                        if not isinstance(format_eval, str):
                            message = (
                                f"'format' expression must evaluate to a string. "
                                f"Got {format_eval!r} from {format_field!r}."
                            )
                            if isinstance(result, list):
                                message += f" 'self' had the value of the index {index} result: {primary!r}."
                            raise SourceLine(schema, "format", WorkflowException, debug).makeError(
                                message
                            )
                        primary["format"] = format_eval
                else:
                    for primary in aslist(result):
                        primary["format"] = format_field
            # Ensure files point to local references outside of the run environment
            adjustFileObjs(result, revmap)

            if not result and optional:
                # Don't convert zero or empty string to None
                if result in [0, ""]:
                    return result
                # For [] or None, return None
                else:
                    return None

        if (
            not result
            and not empty_and_optional
            and isinstance(schema["type"], MutableMapping)
            and schema["type"]["type"] == "record"
        ):
            out = {}
            for field in cast(List[CWLObjectType], schema["type"]["fields"]):
                out[shortname(cast(str, field["name"]))] = self.collect_output(
                    field, builder, outdir, fs_access, compute_checksum=compute_checksum
                )
            return out
        return result



# Stubs for networkx.utils.heaps (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

class MinHeap:
    class _Item:
        key: Any = ...
        value: Any = ...
        def __init__(self, key, value) -> None: ...

    def __init__(self) -> None: ...
    def min(self): ...
    def pop(self): ...
    def get(self, key, default: Optional[Any] = ...): ...
    def insert(self, key, value, allow_increase: bool = ...): ...
    def __nonzero__(self): ...
    def __bool__(self): ...
    def __len__(self): ...
    def __contains__(self, key): ...

class PairingHeap(MinHeap):
    class _Node(MinHeap._Item):
        left: Any = ...
        next: Any = ...
        prev: Any = ...
        parent: Any = ...
        def __init__(self, key, value) -> None: ...

    def __init__(self) -> None: ...
    def min(self): ...
    def pop(self): ...
    def get(self, key, default: Optional[Any] = ...): ...
    def insert(self, key, value, allow_increase: bool = ...): ...

class BinaryHeap(MinHeap):
    def __init__(self) -> None: ...
    def min(self): ...
    def pop(self): ...
    def get(self, key, default: Optional[Any] = ...): ...
    def insert(self, key, value, allow_increase: bool = ...): ...



# Stubs for networkx.algorithms.assortativity.connectivity (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def average_degree_connectivity(
    G,
    source: str = ...,
    target: str = ...,
    nodes: Optional[Any] = ...,
    weight: Optional[Any] = ...,
): ...

k_nearest_neighbors = average_degree_connectivity



# Stubs for networkx.algorithms.approximation (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.algorithms.approximation.clique import *
from networkx.algorithms.approximation.clustering_coefficient import *
from networkx.algorithms.approximation.connectivity import *
from networkx.algorithms.approximation.dominating_set import *
from networkx.algorithms.approximation.independent_set import *
from networkx.algorithms.approximation.kcomponents import *
from networkx.algorithms.approximation.matching import *
from networkx.algorithms.approximation.ramsey import *
from networkx.algorithms.approximation.steinertree import *
from networkx.algorithms.approximation.vertex_cover import *
